# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Various utility functions.
"""

from __future__ import (
    absolute_import,
    print_function,
)

import logging
import os
import socket
import re
import decimal
import dns
import encodings
import sys
import ssl
import termios
import fcntl
import shutil
import struct
import subprocess

import netaddr
from dns import resolver, rdatatype
from dns.exception import DNSException
from dns.resolver import NXDOMAIN
from netaddr.core import AddrFormatError
import six

try:
    from httplib import HTTPSConnection
except ImportError:
    # Python 3
    from http.client import HTTPSConnection

from ipalib import errors, messages
from ipalib.constants import (
    DOMAIN_LEVEL_0,
    TLS_VERSIONS, TLS_VERSION_MINIMAL
)
from ipalib.text import _
from ipaplatform.constants import constants
from ipaplatform.paths import paths
from ipapython.ssh import SSHPublicKey
from ipapython.dn import DN, RDN
from ipapython.dnsutil import DNSName
from ipapython.dnsutil import resolve_ip_addresses
from ipapython.admintool import ScriptError

if sys.version_info >= (3, 2):
    import reprlib  # pylint: disable=import-error
else:
    reprlib = None

if six.PY3:
    unicode = str

_IPA_CLIENT_SYSRESTORE = "/var/lib/ipa-client/sysrestore"
_IPA_DEFAULT_CONF = "/etc/ipa/default.conf"

logger = logging.getLogger(__name__)


def json_serialize(obj):
    if isinstance(obj, (list, tuple)):
        return [json_serialize(o) for o in obj]
    if isinstance(obj, dict):
        return {k: json_serialize(v) for (k, v) in obj.items()}
    if isinstance(obj, (int, bool, float, unicode, type(None))):
        return obj
    if isinstance(obj, str):
        return obj.decode('utf-8')
    if isinstance(obj, (decimal.Decimal, DN)):
        return str(obj)
    if not callable(getattr(obj, '__json__', None)):
        # raise TypeError('%r is not JSON serializable')
        return ''
    return json_serialize(obj.__json__())


def verify_host_resolvable(fqdn):
    try:
        if not resolve_ip_addresses(fqdn):
            raise errors.DNSNotARecordError(hostname=fqdn)
    except dns.exception.DNSException as ex:
        # wrap DNSException in a PublicError
        raise errors.DNSResolverError(exception=ex)


def has_soa_or_ns_record(domain):
    """
    Checks to see if given domain has SOA or NS record.
    Returns True or False.
    """
    try:
        resolver.query(domain, rdatatype.SOA)
        soa_record_found = True
    except DNSException:
        soa_record_found = False

    try:
        resolver.query(domain, rdatatype.NS)
        ns_record_found = True
    except DNSException:
        ns_record_found = False

    return soa_record_found or ns_record_found


def normalize_name(name):
    result = dict()
    components = name.split('@')
    if len(components) == 2:
        result['domain'] = unicode(components[1]).lower()
        result['name'] = unicode(components[0]).lower()
    else:
        components = name.split('\\')
        if len(components) == 2:
            result['flatname'] = unicode(components[0]).lower()
            result['name'] = unicode(components[1]).lower()
        else:
            result['name'] = unicode(name).lower()
    return result

def isvalid_base64(data):
    """
    Validate the incoming data as valid base64 data or not. This is only
    used in the ipalib.Parameters module which expects ``data`` to be unicode.

    The character set must only include of a-z, A-Z, 0-9, + or / and
    be padded with = to be a length divisible by 4 (so only 0-2 =s are
    allowed). Its length must be divisible by 4. White space is
    not significant so it is removed.

    This doesn't guarantee we have a base64-encoded value, just that it
    fits the base64 requirements.
    """

    data = ''.join(data.split())

    if (len(data) % 4 > 0 or
            re.match(r'^[a-zA-Z0-9\+\/]+\={0,2}$', data) is None):
        return False
    else:
        return True


def strip_csr_header(csr):
    """
    Remove the header and footer (and surrounding material) from a CSR.
    """
    headerlen = 40
    s = csr.find(b"-----BEGIN NEW CERTIFICATE REQUEST-----")
    if s == -1:
        headerlen = 36
        s = csr.find(b"-----BEGIN CERTIFICATE REQUEST-----")
    if s >= 0:
        e = csr.find(b"-----END")
        csr = csr[s + headerlen:e]

    return csr


def validate_ipaddr(ipaddr):
    """
    Check to see if the given IP address is a valid IPv4 or IPv6 address.

    Returns True or False
    """
    try:
        socket.inet_pton(socket.AF_INET, ipaddr)
    except socket.error:
        try:
            socket.inet_pton(socket.AF_INET6, ipaddr)
        except socket.error:
            return False
    return True

def check_writable_file(filename):
    """
    Determine if the file is writable. If the file doesn't exist then
    open the file to test writability.
    """
    if filename is None:
        raise errors.FileError(reason=_('Filename is empty'))
    try:
        if os.path.isfile(filename):
            if not os.access(filename, os.W_OK):
                raise errors.FileError(reason=_('Permission denied: %(file)s') % dict(file=filename))
        else:
            fp = open(filename, 'w')
            fp.close()
    except (IOError, OSError) as e:
        raise errors.FileError(reason=str(e))

def normalize_zonemgr(zonemgr):
    if not zonemgr or not isinstance(zonemgr, str):
        return zonemgr
    if '@' in zonemgr:
        # local-part needs to be normalized
        name, _at, domain = zonemgr.partition('@')
        name = name.replace('.', '\\.')
        zonemgr = u''.join((name, u'.', domain))

    return zonemgr

def normalize_zone(zone):
    if zone[-1] != '.':
        return zone + '.'
    else:
        return zone


def get_proper_tls_version_span(tls_version_min, tls_version_max):
    """
    This function checks whether the given TLS versions are known in
    FreeIPA and that these versions fulfill the requirements for minimal
    TLS version (see
    `ipalib.constants: TLS_VERSIONS, TLS_VERSION_MINIMAL`).

    :param tls_version_min:
        the lower value in the TLS min-max span, raised to the lowest
        allowed value if too low
    :param tls_version_max:
        the higher value in the TLS min-max span, raised to tls_version_min
        if lower than TLS_VERSION_MINIMAL
    :raises: ValueError
    """
    min_allowed_idx = TLS_VERSIONS.index(TLS_VERSION_MINIMAL)

    try:
        min_version_idx = TLS_VERSIONS.index(tls_version_min)
    except ValueError:
        raise ValueError("tls_version_min ('{val}') is not a known "
                         "TLS version.".format(val=tls_version_min))

    try:
        max_version_idx = TLS_VERSIONS.index(tls_version_max)
    except ValueError:
        raise ValueError("tls_version_max ('{val}') is not a known "
                         "TLS version.".format(val=tls_version_max))

    if min_version_idx > max_version_idx:
        raise ValueError("tls_version_min is higher than "
                         "tls_version_max.")

    if min_version_idx < min_allowed_idx:
        min_version_idx = min_allowed_idx
        logger.warning("tls_version_min set too low ('%s'),using '%s' instead",
                       tls_version_min, TLS_VERSIONS[min_version_idx])

    if max_version_idx < min_allowed_idx:
        max_version_idx = min_version_idx
        logger.warning("tls_version_max set too low ('%s'),using '%s' instead",
                       tls_version_max, TLS_VERSIONS[max_version_idx])
    return TLS_VERSIONS[min_version_idx:max_version_idx+1]


def create_https_connection(
    host, port=HTTPSConnection.default_port,
    cafile=None,
    client_certfile=None, client_keyfile=None,
    keyfile_passwd=None,
    tls_version_min="tls1.1",
    tls_version_max="tls1.2",
    **kwargs
):
    """
    Create a customized HTTPSConnection object.

    :param host:  The host to connect to
    :param port:  The port to connect to, defaults to
               HTTPSConnection.default_port
    :param cafile:  A PEM-format file containning the trusted
                    CA certificates
    :param client_certfile:
            A PEM-format client certificate file that will be used to
            identificate the user to the server.
    :param client_keyfile:
            A file with the client private key. If this argument is not
            supplied, the key will be sought in client_certfile.
    :param keyfile_passwd:
            A path to the file which stores the password that is used to
            encrypt client_keyfile. Leave default value if the keyfile
            is not encrypted.
    :returns An established HTTPS connection to host:port
    """
    # pylint: disable=no-member
    tls_cutoff_map = {
        "ssl2": ssl.OP_NO_SSLv2,
        "ssl3": ssl.OP_NO_SSLv3,
        "tls1.0": ssl.OP_NO_TLSv1,
        "tls1.1": ssl.OP_NO_TLSv1_1,
        "tls1.2": ssl.OP_NO_TLSv1_2,
    }
    # pylint: enable=no-member

    if cafile is None:
        raise RuntimeError("cafile argument is required to perform server "
                           "certificate verification")

    if not os.path.isfile(cafile) or not os.access(cafile, os.R_OK):
        raise RuntimeError("cafile \'{file}\' doesn't exist or is unreadable".
                           format(file=cafile))

    # remove the slice of negating protocol options according to options
    tls_span = get_proper_tls_version_span(tls_version_min, tls_version_max)

    # official Python documentation states that the best option to get
    # TLSv1 and later is to setup SSLContext with PROTOCOL_SSLv23
    # and then negate the insecure SSLv2 and SSLv3
    # pylint: disable=no-member
    ctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
    ctx.options |= (
        ssl.OP_ALL | ssl.OP_NO_COMPRESSION | ssl.OP_SINGLE_DH_USE |
        ssl.OP_SINGLE_ECDH_USE
    )

    # high ciphers without RC4, MD5, TripleDES, pre-shared key and secure
    # remote password. Uses system crypto policies on some platforms.
    ctx.set_ciphers(constants.TLS_HIGH_CIPHERS)

    # pylint: enable=no-member
    # set up the correct TLS version flags for the SSL context
    for version in TLS_VERSIONS:
        if version in tls_span:
            # make sure the required TLS versions are available if Python
            # decides to modify the default TLS flags
            ctx.options &= ~tls_cutoff_map[version]
        else:
            # disable all TLS versions not in tls_span
            ctx.options |= tls_cutoff_map[version]

    ctx.verify_mode = ssl.CERT_REQUIRED
    ctx.check_hostname = True
    ctx.load_verify_locations(cafile)

    if client_certfile is not None:
        if keyfile_passwd is not None:
            with open(keyfile_passwd) as pwd_f:
                passwd = pwd_f.read()
        else:
            passwd = None
        ctx.load_cert_chain(client_certfile, client_keyfile, passwd)

    return HTTPSConnection(host, port, context=ctx, **kwargs)


def validate_dns_label(dns_label, allow_underscore=False, allow_slash=False):
    base_chars = 'a-z0-9'
    extra_chars = ''
    middle_chars = ''

    if allow_underscore:
        extra_chars += '_'
    if allow_slash:
        middle_chars += '/'

    middle_chars = middle_chars + '-' #has to be always the last in the regex [....-]

    label_regex = r'''^[%(base)s%(extra)s] # must begin with an alphanumeric
                                           # character, or underscore if
                                           # allow_underscore is True
        ([%(base)s%(extra)s%(middle)s]*    # can contain all allowed character
                                           # classes in the middle
        [%(base)s%(extra)s])*$             # must end with alphanumeric
                                           # character or underscore if
                                           # allow_underscore is True
        ''' % dict(base=base_chars, extra=extra_chars, middle=middle_chars)
    regex = re.compile(label_regex, re.IGNORECASE | re.VERBOSE)

    if not dns_label:
        raise ValueError(_('empty DNS label'))

    if len(dns_label) > 63:
        raise ValueError(_('DNS label cannot be longer that 63 characters'))

    if not regex.match(dns_label):
        chars = ', '.join("'%s'" % c for c in extra_chars + middle_chars)
        chars2 = ', '.join("'%s'" % c for c in middle_chars)
        raise ValueError(_("only letters, numbers, %(chars)s are allowed. " \
                           "DNS label may not start or end with %(chars2)s") \
                           % dict(chars=chars, chars2=chars2))


def validate_domain_name(
    domain_name, allow_underscore=False,
    allow_slash=False, entity='domain'
):
    if domain_name.endswith('.'):
        domain_name = domain_name[:-1]

    domain_name = domain_name.split(".")

    if len(domain_name) < 2:
        raise ValueError(_(
            'single label {}s are not supported'.format(entity)))

    # apply DNS name validator to every name part
    for label in domain_name:
        validate_dns_label(label, allow_underscore, allow_slash)


def validate_zonemgr(zonemgr):
    assert isinstance(zonemgr, DNSName)
    if any(b'@' in label for label in zonemgr.labels):
        raise ValueError(_('too many \'@\' characters'))


def validate_zonemgr_str(zonemgr):
    zonemgr = normalize_zonemgr(zonemgr)
    validate_idna_domain(zonemgr)
    zonemgr = DNSName(zonemgr)
    return validate_zonemgr(zonemgr)

def validate_hostname(hostname, check_fqdn=True, allow_underscore=False, allow_slash=False):
    """ See RFC 952, 1123

    :param hostname Checked value
    :param check_fqdn Check if hostname is fully qualified
    """
    if len(hostname) > 255:
        raise ValueError(_('cannot be longer that 255 characters'))

    if hostname.endswith('.'):
        hostname = hostname[:-1]

    if '..' in hostname:
        raise ValueError(_('hostname contains empty label (consecutive dots)'))

    if '.' not in hostname:
        if check_fqdn:
            raise ValueError(_('not fully qualified'))
        validate_dns_label(hostname, allow_underscore, allow_slash)
    else:
        validate_domain_name(hostname, allow_underscore, allow_slash)

def normalize_sshpubkey(value):
    return SSHPublicKey(value).openssh()


def validate_sshpubkey(ugettext, value):
    try:
        SSHPublicKey(value)
    except (ValueError, UnicodeDecodeError):
        return _('invalid SSH public key')
    else:
        return None


def validate_sshpubkey_no_options(ugettext, value):
    try:
        pubkey = SSHPublicKey(value)
    except (ValueError, UnicodeDecodeError):
        return _('invalid SSH public key')

    if pubkey.has_options():
        return _('options are not allowed')
    else:
        return None


def convert_sshpubkey_post(entry_attrs):
    pubkeys = entry_attrs.get('ipasshpubkey')
    if not pubkeys:
        return

    newpubkeys = []
    fingerprints = []
    for pubkey in pubkeys:
        try:
            pubkey = SSHPublicKey(pubkey)
        except (ValueError, UnicodeDecodeError):
            continue

        fp = pubkey.fingerprint_hex_sha256()
        comment = pubkey.comment()
        if comment:
            fp = u'%s %s' % (fp, comment)
        fp = u'%s (%s)' % (fp, pubkey.keytype())

        newpubkeys.append(pubkey.openssh())
        fingerprints.append(fp)

    if 'ipasshpubkey' in entry_attrs:
        entry_attrs['ipasshpubkey'] = newpubkeys or None
    if fingerprints:
        entry_attrs['sshpubkeyfp'] = fingerprints


def add_sshpubkey_to_attrs_pre(context, attrs_list):
    """
    Attribute ipasshpubkey should be added to attrs_list to be able compute
    ssh fingerprint. This attribute must be removed later if was added here
    (see remove_sshpubkey_from_output_post).
    """
    if not ('ipasshpubkey' in attrs_list or '*' in attrs_list):
        setattr(context, 'ipasshpubkey_added', True)
        attrs_list.append('ipasshpubkey')


def remove_sshpubkey_from_output_post(context, entry_attrs):
    """
    Remove ipasshpubkey from output if it was added in pre_callbacks
    """
    if getattr(context, 'ipasshpubkey_added', False):
        entry_attrs.pop('ipasshpubkey', None)
        delattr(context, 'ipasshpubkey_added')


def remove_sshpubkey_from_output_list_post(context, entries):
    """
    Remove ipasshpubkey from output if it was added in pre_callbacks
    """
    if getattr(context, 'ipasshpubkey_added', False):
        for entry_attrs in entries:
            entry_attrs.pop('ipasshpubkey', None)
        delattr(context, 'ipasshpubkey_added')


# regexp matching signed floating point number (group 1) followed by
# optional whitespace followed by time unit, e.g. day, hour (group 7)
time_duration_re = re.compile(r'([-+]?((\d+)|(\d+\.\d+)|(\.\d+)|(\d+\.)))\s*([a-z]+)', re.IGNORECASE)

# number of seconds in a time unit
time_duration_units = {
    'year'    : 365*24*60*60,
    'years'   : 365*24*60*60,
    'y'       : 365*24*60*60,
    'month'   : 30*24*60*60,
    'months'  : 30*24*60*60,
    'week'    : 7*24*60*60,
    'weeks'   : 7*24*60*60,
    'w'       : 7*24*60*60,
    'day'     : 24*60*60,
    'days'    : 24*60*60,
    'd'       : 24*60*60,
    'hour'    : 60*60,
    'hours'   : 60*60,
    'h'       : 60*60,
    'minute'  : 60,
    'minutes' : 60,
    'min'     : 60,
    'second'  : 1,
    'seconds' : 1,
    'sec'     : 1,
    's'       : 1,
}

def parse_time_duration(value):
    '''

    Given a time duration string, parse it and return the total number
    of seconds represented as a floating point value. Negative values
    are permitted.

    The string should be composed of one or more numbers followed by a
    time unit. Whitespace and punctuation is optional. The numbers may
    be optionally signed.  The time units are case insenstive except
    for the single character 'M' or 'm' which means month and minute
    respectively.

    Recognized time units are:

        * year, years, y
        * month, months, M
        * week, weeks, w
        * day, days, d
        * hour, hours, h
        * minute, minutes, min, m
        * second, seconds, sec, s

    Examples:
        "1h"                    # 1 hour
        "2 HOURS, 30 Minutes"   # 2.5 hours
        "1week -1 day"          # 6 days
        ".5day"                 # 12 hours
        "2M"                    # 2 months
        "1h:15m"                # 1.25 hours
        "1h, -15min"            # 45 minutes
        "30 seconds"            # .5 minute

    Note: Despite the appearance you can perform arithmetic the
    parsing is much simpler, the parser searches for signed values and
    adds the signed value to a running total. Only + and - are permitted
    and must appear prior to a digit.

    :parameters:
        value : string
            A time duration string in the specified format
    :returns:
        total number of seconds as float (may be negative)
    '''

    matches = 0
    duration = 0.0
    for match in time_duration_re.finditer(value):
        matches += 1
        magnitude = match.group(1)
        unit = match.group(7)

        # Get the unit, only M and m are case sensitive
        if unit == 'M':         # month
            seconds_per_unit = 30*24*60*60
        elif unit == 'm':       # minute
            seconds_per_unit = 60
        else:
            unit = unit.lower()
            seconds_per_unit = time_duration_units.get(unit)
            if seconds_per_unit is None:
                raise ValueError('unknown time duration unit "%s"' % unit)
        magnitude = float(magnitude)
        seconds = magnitude * seconds_per_unit
        duration += seconds

    if matches == 0:
        raise ValueError('no time duration found in "%s"' % value)

    return duration

def get_dns_forward_zone_update_policy(realm, rrtypes=('A', 'AAAA', 'SSHFP')):
    """
    Generate update policy for a forward DNS zone (idnsUpdatePolicy
    attribute). Bind uses this policy to grant/reject access for client
    machines trying to dynamically update their records.

    :param realm: A realm of the of the client
    :param rrtypes: A list of resource records types that client shall be
                    allowed to update
    """
    policy_element = "grant %(realm)s krb5-self * %(rrtype)s"
    policies = [ policy_element % dict(realm=realm, rrtype=rrtype) \
               for rrtype in rrtypes ]
    policy = "; ".join(policies)
    policy += ";"

    return policy

def get_dns_reverse_zone_update_policy(realm, reverse_zone, rrtypes=('PTR',)):
    """
    Generate update policy for a reverse DNS zone (idnsUpdatePolicy
    attribute). Bind uses this policy to grant/reject access for client
    machines trying to dynamically update their records.

    :param realm: A realm of the of the client
    :param reverse_zone: Name of the actual zone. All clients with IPs in this
                         sub-domain will be allowed to perform changes
    :param rrtypes: A list of resource records types that client shall be
                    allowed to update
    """
    policy_element = "grant %(realm)s krb5-subdomain %(zone)s %(rrtype)s"
    policies = [ policy_element \
                    % dict(realm=realm, zone=reverse_zone, rrtype=rrtype) \
                 for rrtype in rrtypes ]
    policy = "; ".join(policies)
    policy += ";"

    return policy

# dictionary of valid reverse zone -> number of address components
REVERSE_DNS_ZONES = {
    DNSName.ip4_rev_zone : 4,
    DNSName.ip6_rev_zone : 32,
}

def zone_is_reverse(zone_name):
    return DNSName(zone_name).is_reverse()

def get_reverse_zone_default(ip_address):
    ip = netaddr.IPAddress(str(ip_address))
    items = ip.reverse_dns.split('.')

    if ip.version == 4:
        items = items[1:]   # /24 for IPv4
    elif ip.version == 6:
        items = items[16:]  # /64 for IPv6

    return normalize_zone('.'.join(items))


def validate_rdn_param(ugettext, value):
    try:
        RDN(value)
    except Exception as e:
        return str(e)
    else:
        return None


def validate_hostmask(ugettext, hostmask):
    try:
        netaddr.IPNetwork(hostmask)
    except (ValueError, AddrFormatError):
        return _('invalid hostmask')
    else:
        return None


class ForwarderValidationError(Exception):
    format = None

    def __init__(self, format=None, message=None, **kw):
        messages.process_message_arguments(self, format, message, **kw)
        super(ForwarderValidationError, self).__init__(self.msg)


class UnresolvableRecordError(ForwarderValidationError):
    format = _("query '%(owner)s %(rtype)s': %(error)s")


class EDNS0UnsupportedError(ForwarderValidationError):
    format = _("query '%(owner)s %(rtype)s' with EDNS0: %(error)s")


class DNSSECSignatureMissingError(ForwarderValidationError):
    format = _("answer to query '%(owner)s %(rtype)s' is missing DNSSEC "
               "signatures (no RRSIG data)")


class DNSSECValidationError(ForwarderValidationError):
    format = _("record '%(owner)s %(rtype)s' "
               "failed DNSSEC validation on server %(ip)s")


def _log_response(e):
    """
    If exception contains response from server, log this response to debug log
    :param log: if log is None, do not log
    :param e: DNSException
    """
    assert isinstance(e, DNSException)
    response = getattr(e, 'kwargs', {}).get('response')
    if response:
        logger.debug("DNSException: %s; server response: %s", e, response)


def _resolve_record(owner, rtype, nameserver_ip=None, edns0=False,
                    dnssec=False, flag_cd=False, timeout=10):
    """
    :param nameserver_ip: if None, default resolvers will be used
    :param edns0: enables EDNS0
    :param dnssec: enabled EDNS0, flags: DO
    :param flag_cd: requires dnssec=True, adds flag CD
    :raise DNSException: if error occurs
    """
    assert isinstance(nameserver_ip, str) or nameserver_ip is None
    assert isinstance(rtype, str)

    res = dns.resolver.Resolver()
    if nameserver_ip:
        res.nameservers = [nameserver_ip]
    res.lifetime = timeout

    # Recursion Desired,
    # this option prevents to get answers in authority section instead of answer
    res.set_flags(dns.flags.RD)

    if dnssec:
        res.use_edns(0, dns.flags.DO, 4096)
        flags = dns.flags.RD
        if flag_cd:
            flags = flags | dns.flags.CD
        res.set_flags(flags)
    elif edns0:
        res.use_edns(0, 0, 4096)

    return res.query(owner, rtype)


def _validate_edns0_forwarder(owner, rtype, ip_addr, timeout=10):
    """
    Validate if forwarder supports EDNS0

    :raise UnresolvableRecordError: record cannot be resolved
    :raise EDNS0UnsupportedError: EDNS0 is not supported by forwarder
    """

    try:
        _resolve_record(owner, rtype, nameserver_ip=ip_addr, timeout=timeout)
    except DNSException as e:
        _log_response(e)
        raise UnresolvableRecordError(owner=owner, rtype=rtype, ip=ip_addr,
                                      error=e)

    try:
        _resolve_record(owner, rtype, nameserver_ip=ip_addr, edns0=True,
                        timeout=timeout)
    except DNSException as e:
        _log_response(e)
        raise EDNS0UnsupportedError(owner=owner, rtype=rtype, ip=ip_addr,
                                    error=e)


def validate_dnssec_global_forwarder(ip_addr, timeout=10):
    """Test DNS forwarder properties. against root zone.

    Global forwarders should be able return signed root zone

    :raise UnresolvableRecordError: record cannot be resolved
    :raise EDNS0UnsupportedError: EDNS0 is not supported by forwarder
    :raise DNSSECSignatureMissingError: did not receive RRSIG for root zone
    """

    ip_addr = str(ip_addr)
    owner = "."
    rtype = "SOA"

    _validate_edns0_forwarder(owner, rtype, ip_addr, timeout=timeout)

    # DNS root has to be signed
    try:
        ans = _resolve_record(owner, rtype, nameserver_ip=ip_addr, dnssec=True,
                              timeout=timeout)
    except DNSException as e:
        _log_response(e)
        raise DNSSECSignatureMissingError(owner=owner, rtype=rtype, ip=ip_addr)

    try:
        ans.response.find_rrset(
            ans.response.answer, dns.name.root, dns.rdataclass.IN,
            dns.rdatatype.RRSIG, dns.rdatatype.SOA
        )
    except KeyError:
        raise DNSSECSignatureMissingError(owner=owner, rtype=rtype, ip=ip_addr)


def validate_dnssec_zone_forwarder_step1(ip_addr, fwzone, timeout=10):
    """
    Only forwarders in forward zones can be validated in this way
    :raise UnresolvableRecordError: record cannot be resolved
    :raise EDNS0UnsupportedError: ENDS0 is not supported by forwarder
    """
    _validate_edns0_forwarder(fwzone, "SOA", ip_addr, timeout=timeout)


def validate_dnssec_zone_forwarder_step2(ipa_ip_addr, fwzone, timeout=10):
    """
    This step must be executed after forwarders are added into LDAP, and only
    when we are sure the forwarders work.
    Query will be send to IPA DNS server, to verify if reply passed,
    or DNSSEC validation failed.
    Only forwarders in forward zones can be validated in this way
    :raise UnresolvableRecordError: record cannot be resolved
    :raise DNSSECValidationError: response from forwarder is not DNSSEC valid
    """
    rtype = "SOA"
    try:
        ans_cd = _resolve_record(fwzone, rtype, nameserver_ip=ipa_ip_addr,
                                 edns0=True, dnssec=True, flag_cd=True,
                                 timeout=timeout)
    except NXDOMAIN as e:
        # sometimes CD flag is ignored and NXDomain is returned
        _log_response(e)
        raise DNSSECValidationError(owner=fwzone, rtype=rtype, ip=ipa_ip_addr)
    except DNSException as e:
        _log_response(e)
        raise UnresolvableRecordError(owner=fwzone, rtype=rtype,
                                      ip=ipa_ip_addr, error=e)

    try:
        ans_do = _resolve_record(fwzone, rtype, nameserver_ip=ipa_ip_addr,
                                 edns0=True, dnssec=True, timeout=timeout)
    except DNSException as e:
        _log_response(e)
        raise DNSSECValidationError(owner=fwzone, rtype=rtype, ip=ipa_ip_addr)
    else:
        if (ans_do.canonical_name == ans_cd.canonical_name
            and ans_do.rrset == ans_cd.rrset):
            return
        # records received with and without CD flag are not equivalent:
        # this might be caused by an DNSSEC validation failure in cases where
        # existing zone id being 'shadowed' by another zone on forwarder
        raise DNSSECValidationError(owner=fwzone, rtype=rtype, ip=ipa_ip_addr)


def validate_idna_domain(value):
    """
    Validate if value is valid IDNA domain.

    If domain is not valid, raises ValueError
    :param value:
    :return:
    """
    error = None

    try:
        DNSName(value)
    except dns.name.BadEscape:
        error = _('invalid escape code in domain name')
    except dns.name.EmptyLabel:
        error = _('empty DNS label')
    except dns.name.NameTooLong:
        error = _('domain name cannot be longer than 255 characters')
    except dns.name.LabelTooLong:
        error = _('DNS label cannot be longer than 63 characters')
    except dns.exception.SyntaxError:
        error = _('invalid domain name')
    else:
        #compare if IDN normalized and original domain match
        #there is N:1 mapping between unicode and IDNA names
        #user should use normalized names to avoid mistakes
        labels = re.split(u'[.\uff0e\u3002\uff61]', value, flags=re.UNICODE)
        try:
            for label in labels:
                label.encode("ascii")
        except UnicodeError:
            # IDNA
            is_nonnorm = any(encodings.idna.nameprep(x) != x for x in labels)
            if is_nonnorm:
                error = _("domain name '%(domain)s' should be normalized to"
                          ": %(normalized)s") % {
                          'domain': value,
                          'normalized': '.'.join([encodings.idna.nameprep(x)
                                                  for x in labels])}

    if error:
        raise ValueError(error)


def detect_dns_zone_realm_type(api, domain):
    """
    Detects the type of the realm that the given DNS zone belongs to.
    Note: This method is heuristic. Possible values:
      - 'current': For IPA domains belonging in the current realm.
      - 'foreign': For domains belonging in a foreing kerberos realm.
      - 'unknown': For domains whose allegiance could not be detected.
    """

    # First, try to detect _kerberos TXT record in the domain
    # This would indicate that the domain belongs to IPA realm

    kerberos_prefix = DNSName('_kerberos')
    domain_suffix = DNSName(domain)
    kerberos_record_name = kerberos_prefix + domain_suffix

    try:
        result = resolver.query(kerberos_record_name, rdatatype.TXT)
        answer = result.response.answer

        # IPA domain will have only one _kerberos TXT record
        if (len(answer) == 1 and
            len(answer[0]) == 1 and
            answer[0].rdtype == rdatatype.TXT):

            record = answer[0][0]

            # If the record contains our current realm, it is 'ipa-current'
            if record.to_text() == '"{0}"'.format(api.env.realm):
                return 'current'
            else:
                return 'foreign'

    except DNSException:
        pass

    # Try to detect AD specific record in the zone.
    # This would indicate that the domain belongs to foreign (AD) realm

    gc_prefix = DNSName('_ldap._tcp.gc._msdcs')
    ad_specific_record_name = gc_prefix + domain_suffix

    try:
        # The presence of this record is enough, return foreign in such case
        resolver.query(ad_specific_record_name, rdatatype.SRV)
    except DNSException:
        # If we could not detect type with certainty, return unknown
        return 'unknown'
    else:
        return 'foreign'


def has_managed_topology(api):
    domainlevel = api.Command['domainlevel_get']().get('result', DOMAIN_LEVEL_0)
    return domainlevel > DOMAIN_LEVEL_0


class classproperty:
    __slots__ = ('__doc__', 'fget')

    def __init__(self, fget=None, doc=None):
        if doc is None and fget is not None:
            doc = fget.__doc__

        self.fget = fget
        self.__doc__ = doc

    def __get__(self, obj, obj_type):
        if self.fget is not None:
            return self.fget.__get__(obj, obj_type)()
        raise AttributeError("unreadable attribute")

    def __set__(self, obj, value):
        raise AttributeError("can't set attribute")

    def __delete__(self, obj):
        raise AttributeError("can't delete attribute")

    def getter(self, fget):
        self.fget = fget
        return self


def normalize_hostname(hostname):
    """Use common fqdn form without the trailing dot"""
    if hostname.endswith(u'.'):
        hostname = hostname[:-1]
    hostname = hostname.lower()
    return hostname


def hostname_validator(ugettext, value):
    try:
        validate_hostname(value)
    except ValueError as e:
        return _('invalid domain-name: %s') % unicode(e)

    return None


def ipaddr_validator(ugettext, ipaddr, ip_version=None):
    try:
        ip = netaddr.IPAddress(str(ipaddr), flags=netaddr.INET_PTON)

        if ip_version is not None:
            if ip.version != ip_version:
                return _(
                    'invalid IP address version (is %(value)d, must be '
                    '%(required_value)d)!') % dict(
                    value=ip.version,
                    required_value=ip_version
                )
    except (netaddr.AddrFormatError, ValueError):
        return _('invalid IP address format')
    return None


def validate_bind_forwarder(ugettext, forwarder):
    ip_address, sep, port = forwarder.partition(u' port ')

    ip_address_validation = ipaddr_validator(ugettext, ip_address)

    if ip_address_validation is not None:
        return ip_address_validation

    if sep:
        try:
            port = int(port)
            if port < 0 or port > 65535:
                raise ValueError()
        except ValueError:
            return _('%(port)s is not a valid port' % dict(port=port))

    return None


def set_krbcanonicalname(entry_attrs):
    objectclasses = set(i.lower() for i in entry_attrs['objectclass'])

    if 'krbprincipalaux' not in objectclasses:
        return

    if ('krbprincipalname' in entry_attrs
            and 'krbcanonicalname' not in entry_attrs):
        entry_attrs['krbcanonicalname'] = entry_attrs['krbprincipalname']


def ensure_last_krbprincipalname(ldap, entry_attrs, *keys):
    """
    ensure that the LDAP entry has at least one value of krbprincipalname
    and that this value is equal to krbcanonicalname

    :param ldap: LDAP connection object
    :param entry_attrs: LDAP entry made prior to update
    :param options: command options
    """
    entry = ldap.get_entry(
        entry_attrs.dn, ['krbcanonicalname', 'krbprincipalname'])

    krbcanonicalname = entry.single_value.get('krbcanonicalname', None)

    if krbcanonicalname in keys[-1]:
        raise errors.ValidationError(
            name='krbprincipalname',
            error=_('at least one value equal to the canonical '
                    'principal name must be present')
        )


def ensure_krbcanonicalname_set(ldap, entry_attrs):
    old_entry = ldap.get_entry(
        entry_attrs.dn,
        ['krbcanonicalname', 'krbprincipalname', 'objectclass'])

    if old_entry.single_value.get('krbcanonicalname', None) is not None:
        return

    set_krbcanonicalname(old_entry)

    old_entry.pop('krbprincipalname', None)
    old_entry.pop('objectclass', None)

    entry_attrs.update(old_entry)


def check_client_configuration(env=None):
    """
    Check if IPA client is configured on the system.

    Hardcode return code to avoid recursive imports
    """
    if ((env is not None and not os.path.isfile(env.conf_default)) or
       (not os.path.isfile(paths.IPA_DEFAULT_CONF) or
            not os.path.isdir(paths.IPA_CLIENT_SYSRESTORE) or
            not os.listdir(paths.IPA_CLIENT_SYSRESTORE))):
        raise ScriptError('IPA client is not configured on this system',
                          2)  # CLIENT_NOT_CONFIGURED


def check_principal_realm_in_trust_namespace(api_instance, *keys):
    """
    Check that principal name's suffix does not overlap with UPNs and realm
    names of trusted forests.

    :param api_instance: API instance
    :param suffixes: principal suffixes

    :raises: ValidationError if the suffix coincides with realm name, UPN
    suffix or netbios name of trusted domains
    """
    trust_objects = api_instance.Command.trust_find(u'', sizelimit=0)['result']

    trust_suffix_namespace = set()

    for obj in trust_objects:
        nt_suffixes = obj.get('ipantadditionalsuffixes', [])

        trust_suffix_namespace.update(
            set(upn.lower() for upn in nt_suffixes))

        if 'ipantflatname' in obj:
            trust_suffix_namespace.add(obj['ipantflatname'][0].lower())

        trust_suffix_namespace.add(obj['cn'][0].lower())

    for principal in keys[-1]:
        realm = principal.realm
        upn = principal.upn_suffix if principal.is_enterprise else None

        if realm in trust_suffix_namespace or upn in trust_suffix_namespace:
            raise errors.ValidationError(
                name='krbprincipalname',
                error=_('realm or UPN suffix overlaps with trusted domain '
                        'namespace'))


def no_matching_interface_for_ip_address_warning(addr_list):
    for ip in addr_list:
        if not ip.get_matching_interface():
            logger.warning(
                "No network interface matches the IP address %s", ip)
            # fixme: once when loggers will be fixed, we can remove this
            # print
            print(
                "WARNING: No network interface matches the IP address "
                "{}".format(ip),
                file=sys.stderr
            )


def get_terminal_height(fd=1):
    """
    Get current terminal height

    Args:
        fd (int): file descriptor. Default: 1 (stdout)

    Returns:
        int: Terminal height
    """
    try:
        return struct.unpack(
            'hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, b'1234'))[0]
    except (IOError, OSError, struct.error):
        return os.environ.get("LINES", 25)


def get_pager():
    """ Get path to a pager

    :return: path to the file if it exists otherwise None
    :rtype: str or None
    """
    pager = os.environ.get('PAGER', 'less')
    return shutil.which(pager)


def open_in_pager(data, pager):
    """
    Open text data in pager

    Args:
        data (bytes): data to view in pager
        pager (str): path to the pager

    Returns:
        None
    """
    pager_process = subprocess.Popen([pager], stdin=subprocess.PIPE)

    try:
        pager_process.stdin.write(data)
        pager_process.communicate()
    except IOError:
        pass


if reprlib is not None:
    class APIRepr(reprlib.Repr):
        builtin_types = {
            bool, int, float,
            str, bytes,
            dict, tuple, list, set, frozenset,
            type(None),
        }

        def __init__(self):
            super(APIRepr, self).__init__()
            # no limitation
            for k, v in self.__dict__.items():
                if isinstance(v, int):
                    setattr(self, k, sys.maxsize)

        def repr_str(self, x, level):
            """Output with u'' prefix"""
            return 'u' + repr(x)

        def repr_type(self, x, level):
            if x is str:
                return "<type 'unicode'>"
            if x in self.builtin_types:
                return "<type '{}'>".format(x.__name__)
            else:
                return repr(x)

    apirepr = APIRepr().repr
else:
    apirepr = repr



#
# Copyright (C) 2017  FreeIPA Contributors see COPYING for license
#
from __future__ import absolute_import

import os
import sys

import pytest

import ipaplatform.constants
import ipaplatform.paths
import ipaplatform.services
import ipaplatform.tasks
from ipaplatform._importhook import metaimporter
from ipaplatform.osinfo import osinfo, _parse_osrelease
try:
    from ipaplatform.override import OVERRIDE
except ImportError:
    OVERRIDE = None


HERE = os.path.dirname(os.path.abspath(__file__))
DATA = os.path.join(HERE, 'data')


@pytest.mark.skipif(OVERRIDE is None,
                    reason='test requires override')
def test_override():
    assert OVERRIDE == osinfo.platform_ids[0]
    assert OVERRIDE == osinfo.platform


@pytest.mark.parametrize('mod, name', [
    (ipaplatform.constants, 'ipaplatform.constants'),
    (ipaplatform.paths, 'ipaplatform.paths'),
    (ipaplatform.services, 'ipaplatform.services'),
    (ipaplatform.tasks, 'ipaplatform.tasks'),
])
def test_importhook(mod, name):
    assert name in metaimporter.modules
    prefix, suffix = name.split('.')
    assert prefix == 'ipaplatform'
    override = '.'.join((prefix, metaimporter.platform, suffix))
    assert mod.__name__ == override
    # dicts are equal, modules may not be identical
    assert mod.__dict__ == sys.modules[override].__dict__


@pytest.mark.parametrize('filename, id_, id_like', [
    (os.path.join(DATA, 'os-release-centos'), 'centos', ('rhel', 'fedora')),
    (os.path.join(DATA, 'os-release-fedora'), 'fedora', ()),
    (os.path.join(DATA, 'os-release-ubuntu'), 'ubuntu', ('debian',)),
])
def test_parse_os_release(filename, id_, id_like):
    parsed = _parse_osrelease(filename)
    assert parsed['ID'] == id_
    assert parsed['ID_LIKE'] == id_like



# Authors:
#   Petr Viktorin <pviktori@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import pytest

from ipapython.dn import DN
from ipatests.pytest_ipa.integration import tasks
from ipatests.test_integration.base import IntegrationTest


def check_replication(source_host, dest_host, login):
    source_host.run_command([
        "ipa", "user-add", login,
        "--first", "test",
        "--last", "user"
    ])

    source_ldap = source_host.ldap_connect()
    tasks.wait_for_replication(source_ldap)

    ldap = dest_host.ldap_connect()
    tasks.wait_for_replication(ldap)

    # Check using LDAP
    basedn = dest_host.domain.basedn
    user_dn = DN(
        ("uid", login), ("cn", "users"),
        ("cn", "accounts"), basedn
    )
    entry = ldap.get_entry(user_dn)
    assert entry.dn == user_dn
    assert entry["uid"] == [login]

    # Check using CLI
    result = dest_host.run_command(['ipa', 'user-show', login])
    assert "User login: {}".format(login) in result.stdout_text


@pytest.mark.ds_acceptance
class TestSimpleReplication(IntegrationTest):
    """Simple replication test

    Install a server and a replica, then add an user on one host and ensure
    it is also present on the other one.
    """
    num_replicas = 1
    topology = 'star'

    def test_user_replication_to_replica(self):
        """Test user replication master -> replica"""
        check_replication(self.master, self.replicas[0], 'testuser1')

    def test_user_replication_to_master(self):
        """Test user replication replica -> master"""
        check_replication(self.replicas[0], self.master, 'testuser2')

    def test_replica_removal(self):
        """Test replica removal"""
        result = self.master.run_command(['ipa-replica-manage', 'list'])
        assert self.replicas[0].hostname in result.stdout_text
        # has to be run with --force, there is no --unattended
        self.master.run_command(['ipa-replica-manage', 'del',
                                 self.replicas[0].hostname, '--force'])
        result = self.master.run_command(['ipa-replica-manage', 'list'])
        assert self.replicas[0].hostname not in result.stdout_text



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

from ipapython.install import cli
from ipapython.install.core import extend_knob
from ipaplatform.paths import paths
from ipaserver.install.server import ServerMasterInstall


class CompatServerMasterInstall(ServerMasterInstall):
    all_ip_addresses = False
    nisdomain = None
    no_nisdomain = False
    no_sudo = False
    request_cert = False

    dm_password = extend_knob(
        ServerMasterInstall.dm_password,    # pylint: disable=no-member
        cli_names=['--ds-password', '-p'],
    )

    admin_password = ServerMasterInstall.admin_password
    admin_password = extend_knob(
        admin_password,
        # pylint: disable=no-member
        cli_names=list(admin_password.cli_names) + ['-a'],
    )

    ip_addresses = extend_knob(
        ServerMasterInstall.ip_addresses,   # pylint: disable=no-member
        description="Master Server IP Address. This option can be used "
                    "multiple times",
    )


ServerInstall = cli.install_tool(
    CompatServerMasterInstall,
    command_name='ipa-server-install',
    log_file_name=paths.IPASERVER_INSTALL_LOG,
    console_format='%(message)s',
    debug_option=True,
    verbose=True,
    uninstall_log_file_name=paths.IPASERVER_UNINSTALL_LOG,
)


def run():
    ServerInstall.run_cli()



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#   Martin Kosek <mkosek@redhat.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipalib import _, ngettext
from ipalib import Str
from ipalib import api, crud
from ipalib import output
from ipalib import Object
from ipalib.plugable import Registry
from .baseldap import gen_pkey_only_option, pkey_to_value

__doc__ = _("""
Group to Group Delegation

A permission enables fine-grained delegation of permissions. Access Control
Rules, or instructions (ACIs), grant permission to permissions to perform
given tasks such as adding a user, modifying a group, etc.

Group to Group Delegations grants the members of one group to update a set
of attributes of members of another group.

EXAMPLES:

 Add a delegation rule to allow managers to edit employee's addresses:
   ipa delegation-add --attrs=street --group=managers --membergroup=employees "managers edit employees' street"

 When managing the list of attributes you need to include all attributes
 in the list, including existing ones. Add postalCode to the list:
   ipa delegation-mod --attrs=street --attrs=postalCode --group=managers --membergroup=employees "managers edit employees' street"

 Display our updated rule:
   ipa delegation-show "managers edit employees' street"

 Delete a rule:
   ipa delegation-del "managers edit employees' street"
""")

register = Registry()

ACI_PREFIX=u"delegation"


@register()
class delegation(Object):
    """
    Delegation object.
    """

    bindable = False
    object_name = _('delegation')
    object_name_plural = _('delegations')
    label = _('Delegations')
    label_singular = _('Delegation')

    takes_params = (
        Str('aciname',
            cli_name='name',
            label=_('Delegation name'),
            doc=_('Delegation name'),
            primary_key=True,
        ),
        Str('permissions*',
            cli_name='permissions',
            label=_('Permissions'),
            doc=_('Permissions to grant (read, write). Default is write.'),
        ),
        Str('attrs+',
            cli_name='attrs',
            label=_('Attributes'),
            doc=_('Attributes to which the delegation applies'),
            normalizer=lambda value: value.lower(),
        ),
        Str('memberof',
            cli_name='membergroup',
            label=_('Member user group'),
            doc=_('User group to apply delegation to'),
        ),
        Str('group',
            cli_name='group',
            label=_('User group'),
            doc=_('User group ACI grants access to'),
        ),
        Str('aci',
            label=_('ACI'),
            flags={'no_create', 'no_update', 'no_search'},
        ),
    )

    def __json__(self):
        json_friendly_attributes = (
            'label', 'label_singular', 'takes_params', 'bindable', 'name',
            'object_name', 'object_name_plural',
        )
        json_dict = dict(
            (a, getattr(self, a)) for a in json_friendly_attributes
        )
        json_dict['primary_key'] = self.primary_key.name

        json_dict['methods'] = [m for m in self.methods]
        return json_dict

    def postprocess_result(self, result):
        try:
            # do not include prefix in result
            del result['aciprefix']
        except KeyError:
            pass



@register()
class delegation_add(crud.Create):
    __doc__ = _('Add a new delegation.')

    msg_summary = _('Added delegation "%(value)s"')

    def execute(self, aciname, **kw):
        if 'permissions' not in kw:
            kw['permissions'] = (u'write',)
        kw['aciprefix'] = ACI_PREFIX
        result = api.Command['aci_add'](aciname, **kw)['result']
        self.obj.postprocess_result(result)

        return dict(
            result=result,
            value=pkey_to_value(aciname, kw),
        )



@register()
class delegation_del(crud.Delete):
    __doc__ = _('Delete a delegation.')

    has_output = output.standard_boolean
    msg_summary = _('Deleted delegation "%(value)s"')

    def execute(self, aciname, **kw):
        kw['aciprefix'] = ACI_PREFIX
        result = api.Command['aci_del'](aciname, **kw)
        self.obj.postprocess_result(result)
        return dict(
            result=True,
            value=pkey_to_value(aciname, kw),
        )



@register()
class delegation_mod(crud.Update):
    __doc__ = _('Modify a delegation.')

    msg_summary = _('Modified delegation "%(value)s"')

    def execute(self, aciname, **kw):
        kw['aciprefix'] = ACI_PREFIX
        result = api.Command['aci_mod'](aciname, **kw)['result']
        self.obj.postprocess_result(result)

        return dict(
            result=result,
            value=pkey_to_value(aciname, kw),
        )



@register()
class delegation_find(crud.Search):
    __doc__ = _('Search for delegations.')

    msg_summary = ngettext(
        '%(count)d delegation matched', '%(count)d delegations matched', 0
    )

    takes_options = (gen_pkey_only_option("name"),)

    def execute(self, term=None, **kw):
        kw['aciprefix'] = ACI_PREFIX
        results = api.Command['aci_find'](term, **kw)['result']

        for aci in results:
            self.obj.postprocess_result(aci)

        return dict(
            result=results,
            count=len(results),
            truncated=False,
        )



@register()
class delegation_show(crud.Retrieve):
    __doc__ = _('Display information about a delegation.')

    def execute(self, aciname, **kw):
        result = api.Command['aci_show'](aciname, aciprefix=ACI_PREFIX, **kw)['result']
        self.obj.postprocess_result(result)
        return dict(
            result=result,
            value=pkey_to_value(aciname, kw),
        )



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

from ipaclient.install import client
from ipaplatform.paths import paths
from ipapython.install import cli
from ipapython.install.core import knob, extend_knob


class StandaloneClientInstall(client.ClientInstall):
    no_host_dns = False
    no_wait_for_dns = False

    principal = client.ClientInstall.principal
    principal = extend_knob(
        principal,
        cli_names=list(principal.cli_names) + ['-p'],
    )

    password = knob(
        str, None,
        sensitive=True,
        description="password to join the IPA realm (assumes bulk password "
                    "unless principal is also set)",
        cli_names=[None, '-w'],
    )

    @property
    def admin_password(self):
        if self.principal:
            return self.password

        return super(StandaloneClientInstall, self).admin_password

    @property
    def host_password(self):
        if not self.principal:
            return self.password

        return super(StandaloneClientInstall, self).host_password

    prompt_password = knob(
        None,
        description="Prompt for a password to join the IPA realm",
        cli_names='-W',
    )

    on_master = knob(
        None,
        deprecated=True,
    )


ClientInstall = cli.install_tool(
    StandaloneClientInstall,
    command_name='ipa-client-install',
    log_file_name=paths.IPACLIENT_INSTALL_LOG,
    debug_option=True,
    verbose=True,
    console_format='%(message)s',
    uninstall_log_file_name=paths.IPACLIENT_UNINSTALL_LOG,
    ignore_return_codes=(client.CLIENT_NOT_CONFIGURED,),
)


def run():
    ClientInstall.run_cli()



# Authors:
#     Alexander Bokovoy <abokovoy@redhat.com>
#     Martin Kosek <mkosek@redhat.com>
#
# Copyright (C) 2011  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import six

from ipalib.messages import (
    add_message,
    BrokenTrust)
from ipalib.plugable import Registry
from .baseldap import (
    pkey_to_value,
    entry_to_dict,
    LDAPCreate,
    LDAPDelete,
    LDAPUpdate,
    LDAPSearch,
    LDAPRetrieve,
    LDAPObject,
    LDAPQuery)
from .dns import dns_container_exists
from ipapython.dn import DN
from ipapython.ipautil import realm_to_suffix
from ipalib import api, Str, StrEnum, Password, Bool, _, ngettext, Int, Flag
from ipalib import Command
from ipalib import errors
from ipalib import output
from ldap import SCOPE_SUBTREE
from time import sleep
from ipaserver.dcerpc_common import (TRUST_ONEWAY,
                                     TRUST_BIDIRECTIONAL,
                                     TRUST_JOIN_EXTERNAL,
                                     LSA_TRUST_ATTRIBUTE_NON_TRANSITIVE,
                                     trust_type_string,
                                     trust_direction_string,
                                     trust_status_string)

if six.PY3:
    unicode = str

try:
    import pysss_murmur
    _murmur_installed = True
except Exception as e:
    _murmur_installed = False

try:
    import pysss_nss_idmap
    _nss_idmap_installed = True
except Exception as e:
    _nss_idmap_installed = False

if api.env.in_server and api.env.context in ['lite', 'server']:
    try:
        import ipaserver.dcerpc
        import dbus
        import dbus.mainloop.glib
        _bindings_installed = True
    except ImportError:
        _bindings_installed = False

__doc__ = _("""
Cross-realm trusts

Manage trust relationship between IPA and Active Directory domains.

In order to allow users from a remote domain to access resources in IPA domain,
trust relationship needs to be established. Currently IPA supports only trusts
between IPA and Active Directory domains under control of Windows Server 2008
or later, with functional level 2008 or later.

Please note that DNS on both IPA and Active Directory domain sides should be
configured properly to discover each other. Trust relationship relies on
ability to discover special resources in the other domain via DNS records.

Examples:

1. Establish cross-realm trust with Active Directory using AD administrator
   credentials:

   ipa trust-add --type=ad <ad.domain> --admin <AD domain administrator> \
           --password

2. List all existing trust relationships:

   ipa trust-find

3. Show details of the specific trust relationship:

   ipa trust-show <ad.domain>

4. Delete existing trust relationship:

   ipa trust-del <ad.domain>

Once trust relationship is established, remote users will need to be mapped
to local POSIX groups in order to actually use IPA resources. The mapping
should be done via use of external membership of non-POSIX group and then
this group should be included into one of local POSIX groups.

Example:

1. Create group for the trusted domain admins' mapping and their local POSIX
group:

   ipa group-add --desc='<ad.domain> admins external map' \
           ad_admins_external --external
   ipa group-add --desc='<ad.domain> admins' ad_admins

2. Add security identifier of Domain Admins of the <ad.domain> to the
   ad_admins_external group:

   ipa group-add-member ad_admins_external --external 'AD\\Domain Admins'

3. Allow members of ad_admins_external group to be associated with
   ad_admins POSIX group:

   ipa group-add-member ad_admins --groups ad_admins_external

4. List members of external members of ad_admins_external group to see
   their SIDs:

   ipa group-show ad_admins_external


GLOBAL TRUST CONFIGURATION

When IPA AD trust subpackage is installed and ipa-adtrust-install is run, a
local domain configuration (SID, GUID, NetBIOS name) is generated. These
identifiers are then used when communicating with a trusted domain of the
particular type.

1. Show global trust configuration for Active Directory type of trusts:

   ipa trustconfig-show --type ad

2. Modify global configuration for all trusts of Active Directory type and set
   a different fallback primary group (fallback primary group GID is used as a
   primary user GID if user authenticating to IPA domain does not have any
   other primary GID already set):

   ipa trustconfig-mod --type ad --fallback-primary-group "another AD group"

3. Change primary fallback group back to default hidden group (any group with
   posixGroup object class is allowed):

   ipa trustconfig-mod --type ad --fallback-primary-group "Default SMB Group"
""")

logger = logging.getLogger(__name__)

register = Registry()

_trust_type_option = StrEnum(
         'trust_type',
         cli_name='type',
         label=_('Trust type (ad for Active Directory, default)'),
         values=(u'ad',),
         default=u'ad',
         autofill=True,
        )

DEFAULT_RANGE_SIZE = 200000

DBUS_IFACE_TRUST = 'com.redhat.idm.trust'

CRED_STYLE_SAMBA = 1
CRED_STYLE_KERBEROS = 2


def make_trust_dn(env, trust_type, dn):
    assert isinstance(dn, DN)
    if trust_type:
        container_dn = DN(('cn', trust_type), env.container_trusts, env.basedn)
        return DN(dn, container_dn)
    return dn


def find_adtrust_masters(ldap, api):
    """
    Returns a list of names of IPA servers with ADTRUST component configured.
    """

    try:
        entries, _truncated = ldap.find_entries(
                "cn=ADTRUST",
                base_dn=api.env.container_masters + api.env.basedn
        )
    except errors.NotFound:
        entries = []

    return [entry.dn[1].value for entry in entries]


def verify_samba_component_presence(ldap, api):
    """
    Verifies that Samba is installed and configured on this particular master.
    If Samba is not available, provide a heplful hint with the list of masters
    capable of running the commands.
    """

    adtrust_present = api.Command['adtrust_is_enabled']()['result']

    hint = _(
        ' Alternatively, following servers are capable of running this '
        'command: %(masters)s'
        )

    def raise_missing_component_error(message):
        masters_with_adtrust = find_adtrust_masters(ldap, api)

        # If there are any masters capable of running Samba requiring commands
        # let's advertise them directly
        if masters_with_adtrust:
            message += hint % dict(masters=', '.join(masters_with_adtrust))

        raise errors.NotFound(
            name=_('AD Trust setup'),
            reason=message,
        )

    # We're ok in this case, bail out
    if adtrust_present and _bindings_installed:
        return

    # First check for packages missing
    elif not _bindings_installed:
        error_message = _(
            'Cannot perform the selected command without Samba 4 support '
            'installed. Make sure you have installed server-trust-ad '
            'sub-package of IPA.'
        )

        raise_missing_component_error(error_message)

    # Packages present, but ADTRUST instance is not configured
    elif not adtrust_present:
        error_message = _(
            'Cannot perform the selected command without Samba 4 instance '
            'configured on this machine. Make sure you have run '
            'ipa-adtrust-install on this server.'
        )

        raise_missing_component_error(error_message)


def generate_creds(trustinstance, style, **options):
    """
    Generate string representing credentials using trust instance
    Input:
       trustinstance -- ipaserver.dcerpc.TrustInstance object
       style         -- style of credentials
                        CRED_STYLE_SAMBA -- for using with Samba bindings
                        CRED_STYLE_KERBEROS -- for obtaining Kerberos ticket
       **options     -- options with realm_admin and realm_passwd keys

    Result:
       a string representing credentials with first % separating
       username and password
       None is returned if realm_passwd key returns nothing from options
    """
    creds = None
    password = options.get('realm_passwd', None)
    if password:
        admin_name = options.get('realm_admin')
        sp = []
        sep = '@'
        if style == CRED_STYLE_SAMBA:
            sep = "\\"
            sp = admin_name.split(sep)
            if len(sp) == 1:
                sp.insert(0, trustinstance.remote_domain.info['name'])
        elif style == CRED_STYLE_KERBEROS:
            sp = admin_name.split('\\')
            if len(sp) > 1:
                sp = [sp[1]]
            else:
                sp = admin_name.split(sep)
            if len(sp) == 1:
                sp.append(
                    trustinstance.remote_domain.info['dns_domain'].upper()
                )
        creds = u"{name}%{password}".format(name=sep.join(sp),
                                            password=password)
    return creds


def add_range(myapi, trustinstance, range_name, dom_sid, *keys, **options):
    """
    First, we try to derive the parameters of the ID range based on the
    information contained in the Active Directory.

    If that was not successful, we go for our usual defaults (random base,
    range size 200 000, ipa-ad-trust range type).

    Any of these can be overridden by passing appropriate CLI options
    to the trust-add command.
    """

    range_size = None
    range_type = None
    base_id = None

    # First, get information about ID space from AD
    # However, we skip this step if other than ipa-ad-trust-posix
    # range type is enforced

    if options.get('range_type', None) in (None, u'ipa-ad-trust-posix'):

        # Get the base dn
        domain = keys[-1]
        basedn = realm_to_suffix(domain)

        # Search for information contained in
        # CN=ypservers,CN=ypServ30,CN=RpcServices,CN=System
        info_filter = '(objectClass=msSFU30DomainInfo)'
        info_dn = DN('CN=ypservers,CN=ypServ30,CN=RpcServices,CN=System')\
            + basedn

        # Get the domain validator
        domain_validator = ipaserver.dcerpc.DomainValidator(myapi)
        if not domain_validator.is_configured():
            raise errors.NotFound(
                reason=_('Cannot search in trusted domains without own '
                         'domain configured. Make sure you have run '
                         'ipa-adtrust-install on the IPA server first'))

        creds = None
        if trustinstance:
            # Re-use AD administrator credentials if they were provided
            creds = generate_creds(trustinstance,
                                   style=CRED_STYLE_KERBEROS, **options)
            if creds:
                domain_validator._admin_creds = creds
        # KDC might not get refreshed data at the first time,
        # retry several times
        for _retry in range(10):
            info_list = domain_validator.search_in_dc(domain,
                                                      info_filter,
                                                      None,
                                                      SCOPE_SUBTREE,
                                                      basedn=info_dn,
                                                      quiet=True)

            if info_list:
                info = info_list[0]
                break
            else:
                sleep(2)

        required_msSFU_attrs = ['msSFU30MaxUidNumber', 'msSFU30OrderNumber']

        if not info_list:
            # We were unable to gain UNIX specific info from the AD
            logger.debug("Unable to gain POSIX info from the AD")
        else:
            if all(attr in info for attr in required_msSFU_attrs):
                logger.debug("Able to gain POSIX info from the AD")
                range_type = u'ipa-ad-trust-posix'

                max_uid = info.get('msSFU30MaxUidNumber')
                max_gid = info.get('msSFU30MaxGidNumber', None)
                max_id = int(max(max_uid, max_gid)[0])

                base_id = int(info.get('msSFU30OrderNumber')[0])
                range_size = (1 + (max_id - base_id) // DEFAULT_RANGE_SIZE)\
                    * DEFAULT_RANGE_SIZE

    # Second, options given via the CLI options take precedence to discovery
    if options.get('range_type', None):
        range_type = options.get('range_type', None)
    elif not range_type:
        range_type = u'ipa-ad-trust'

    if options.get('range_size', None):
        range_size = options.get('range_size', None)
    elif not range_size:
        range_size = DEFAULT_RANGE_SIZE

    if options.get('base_id', None):
        base_id = options.get('base_id', None)
    elif not base_id:
        # Generate random base_id if not discovered nor given via CLI
        base_id = DEFAULT_RANGE_SIZE + (
            pysss_murmur.murmurhash3(
                dom_sid,
                len(dom_sid), 0xdeadbeef
            ) % 10000
        ) * DEFAULT_RANGE_SIZE

    # Finally, add new ID range
    myapi.Command['idrange_add'](range_name,
                                 ipabaseid=base_id,
                                 ipaidrangesize=range_size,
                                 ipabaserid=0,
                                 iparangetype=range_type,
                                 ipanttrusteddomainsid=dom_sid)

    # Return the values that were generated inside this function
    return range_type, range_size, base_id


def fetch_trusted_domains_over_dbus(myapi, *keys, **options):
    if not _bindings_installed:
        return

    forest_name = keys[0]
    method_options = []
    if 'realm_server' in options:
        method_options.extend(['--server', options['realm_server']])
    if 'realm_admin' in options:
        method_options.extend(['--admin', options['realm_admin']])
    if 'realm_passwd' in options:
        method_options.extend(['--password', options['realm_passwd']])

    # Calling oddjobd-activated service via DBus has some quirks:
    # - Oddjobd registers multiple canonical names on the same address
    # - python-dbus only follows name owner changes when mainloop is in use
    # See https://fedorahosted.org/oddjob/ticket/2 for details
    dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
    try:
        _ret = 0
        _stdout = ''
        _stderr = ''
        bus = dbus.SystemBus()
        intf = bus.get_object(DBUS_IFACE_TRUST, "/",
                              follow_name_owner_changes=True)
        fetch_domains_method = intf.get_dbus_method(
                'fetch_domains',
                dbus_interface=DBUS_IFACE_TRUST)
        # Oddjobd D-BUS method definition only accepts fixed number
        # of arguments on the command line. Thus, we need to pass
        # remaining ones as ''. There are 30 slots to allow for extension
        # and the number comes from the 'arguments' definition in
        # install/oddjob/etc/oddjobd.conf.d/oddjobd-ipa-trust.conf
        method_arguments = [forest_name]
        method_arguments.extend(method_options)
        method_arguments.extend([''] * (30 - len(method_arguments)))
        (_ret, _stdout, _stderr) = fetch_domains_method(*method_arguments)
    except dbus.DBusException as e:
        logger.error('Failed to call %s.fetch_domains helper.'
                     'DBus exception is %s.', DBUS_IFACE_TRUST, str(e))
        if _ret != 0:
            logger.error('Helper was called for forest %s, return code is %d',
                         forest_name, _ret)
            logger.error('Standard output from the helper:\n%s---\n', _stdout)
            logger.error('Error output from the helper:\n%s--\n', _stderr)
        raise errors.ServerCommandError(
            server=myapi.env.host,
            error=_('Fetching domains from trusted forest failed. '
                    'See details in the error_log')
        )
    return


@register()
class trust(LDAPObject):
    """
    Trust object.
    """
    trust_types = ('ad', 'ipa')
    container_dn = api.env.container_trusts
    object_name = _('trust')
    object_name_plural = _('trusts')
    object_class = ['ipaNTTrustedDomain', 'ipaIDObject']
    default_attributes = ['cn', 'ipantflatname', 'ipanttrusteddomainsid',
                          'ipanttrusttype', 'ipanttrustattributes',
                          'ipanttrustdirection', 'ipanttrustpartner',
                          'ipanttrustforesttrustinfo',
                          'ipanttrustposixoffset',
                          'ipantsupportedencryptiontypes',
                          'ipantadditionalsuffixes']
    search_display_attributes = ['cn', 'ipantflatname',
                                 'ipanttrusteddomainsid', 'ipanttrusttype',
                                 'ipanttrustattributes',
                                 'ipantadditionalsuffixes']
    managed_permissions = {
        'System: Read Trust Information': {
            # Allow reading of attributes needed for SSSD subdomains support
            'non_object': True,
            'ipapermlocation': DN(container_dn, api.env.basedn),
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'cn', 'objectclass',
                'ipantflatname', 'ipantsecurityidentifier',
                'ipanttrusteddomainsid', 'ipanttrustpartner',
                'ipantsidblacklistincoming', 'ipantsidblacklistoutgoing',
                'ipanttrustdirection', 'ipantadditionalsuffixes',
            },
        },

        'System: Read system trust accounts': {
            'non_object': True,
            'ipapermlocation': DN(container_dn, api.env.basedn),
            'replaces_global_anonymous_aci': True,
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'uidnumber', 'gidnumber', 'krbprincipalname'
            },
            'default_privileges': {'ADTrust Agents'},
        },
    }

    label = _('Trusts')
    label_singular = _('Trust')

    takes_params = (
        Str('cn',
            cli_name='realm',
            label=_('Realm name'),
            primary_key=True,
        ),
        Str('ipantflatname',
            cli_name='flat_name',
            label=_('Domain NetBIOS name'),
            flags=['no_create', 'no_update']),
        Str('ipanttrusteddomainsid',
            cli_name='sid',
            label=_('Domain Security Identifier'),
            flags=['no_create', 'no_update']),
        Str('ipantsidblacklistincoming*',
            cli_name='sid_blacklist_incoming',
            label=_('SID blacklist incoming'),
            flags=['no_create']),
        Str('ipantsidblacklistoutgoing*',
            cli_name='sid_blacklist_outgoing',
            label=_('SID blacklist outgoing'),
            flags=['no_create']),
        Str('trustdirection',
            label=_('Trust direction'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
        Str('trusttype',
            label=_('Trust type'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
        Str('truststatus',
            label=_('Trust status'),
            flags={'virtual_attribute', 'no_create', 'no_update', 'no_search'},
        ),
        Str('ipantadditionalsuffixes*',
            cli_name='upn_suffixes',
            label=_('UPN suffixes'),
            flags={'no_create', 'no_search'},
        ),
    )

    def validate_sid_blacklists(self, entry_attrs):
        if not _bindings_installed:
            # SID validator is not available, return
            # Even if invalid SID gets in the trust entry, it won't crash
            # the validation process as it is translated to SID S-0-0
            return
        for attr in ('ipantsidblacklistincoming', 'ipantsidblacklistoutgoing'):
            values = entry_attrs.get(attr)
            if not values:
                continue
            for value in values:
                if not ipaserver.dcerpc.is_sid_valid(value):
                    err = _("invalid SID: {SID}").format(SID=value)
                    raise errors.ValidationError(name=attr, error=err)

    def get_dn(self, *keys, **kwargs):
        trust_type = kwargs.get('trust_type')

        sdn = [('cn', x) for x in keys]
        sdn.reverse()

        if trust_type is None:
            ldap = self.backend
            trustfilter = ldap.make_filter({
                'objectclass': ['ipaNTTrustedDomain', 'ipaIDObject'],
                'cn': [keys[-1]]},
                rules=ldap.MATCH_ALL
            )

            # more type of objects can be located in subtree (for example
            # cross-realm principals). we need this attr do detect trust
            # entries
            trustfilter = ldap.combine_filters(
                (trustfilter, "ipaNTTrustPartner=*"),
                rules=ldap.MATCH_ALL
            )

            try:
                result = ldap.get_entries(
                    DN(self.container_dn, self.env.basedn),
                    ldap.SCOPE_SUBTREE, trustfilter, ['']
                )
            except errors.NotFound:
                raise self.handle_not_found(keys[-1])

            if len(result) > 1:
                raise errors.OnlyOneValueAllowed(attr='trust domain')

            return result[0].dn

        return make_trust_dn(self.env, trust_type, DN(*sdn))

    def warning_if_ad_trust_dom_have_missing_SID(self, result, **options):
        """Due bug https://fedorahosted.org/freeipa/ticket/5665 there might be
        AD trust domain without generated SID, warn user about it.
        """
        ldap = self.api.Backend.ldap2

        try:
            entries, _truncated = ldap.find_entries(
                base_dn=DN(self.api.env.container_adtrusts,
                           self.api.env.basedn),
                scope=ldap.SCOPE_ONELEVEL,
                attrs_list=['cn'],
                filter='(&(ipaNTTrustPartner=*)'
                       '(!(ipaNTSecurityIdentifier=*)))',
            )
        except errors.NotFound:
            pass
        else:
            for entry in entries:
                add_message(
                    options['version'],
                    result,
                    BrokenTrust(domain=entry.single_value['cn'])
                )


@register()
class trust_add(LDAPCreate):
    __doc__ = _('''
Add new trust to use.

This command establishes trust relationship to another domain
which becomes 'trusted'. As result, users of the trusted domain
may access resources of this domain.

Only trusts to Active Directory domains are supported right now.

The command can be safely run multiple times against the same domain,
this will cause change to trust relationship credentials on both
sides.

Note that if the command was previously run with a specific range type,
or with automatic detection of the range type, and you want to configure a
different range type, you may need to delete first the ID range using
ipa idrange-del before retrying the command with the desired range type.
    ''')

    range_types = {
        u'ipa-ad-trust': unicode(_('Active Directory domain range')),
        u'ipa-ad-trust-posix': unicode(_('Active Directory trust range with '
                                         'POSIX attributes')),
                  }

    takes_options = LDAPCreate.takes_options + (
        _trust_type_option,
        Str('realm_admin?',
            cli_name='admin',
            label=_("Active Directory domain administrator"),
            ),
        Password('realm_passwd?',
                 cli_name='password',
                 label=_("Active Directory domain administrator's password"),
                 confirm=False,
                 ),
        Str('realm_server?',
            cli_name='server',
            label=_('Domain controller for the Active Directory domain '
                    '(optional)'),
            ),
        Password('trust_secret?',
                 cli_name='trust_secret',
                 label=_('Shared secret for the trust'),
                 confirm=False,
                 ),
        Int('base_id?',
            cli_name='base_id',
            label=_('First Posix ID of the range reserved for the '
                    'trusted domain'),
            ),
        Int('range_size?',
            cli_name='range_size',
            label=_('Size of the ID range reserved for the trusted domain')
            ),
        StrEnum('range_type?',
                label=_('Range type'),
                cli_name='range_type',
                doc=_('Type of trusted domain ID range, one of allowed ' +
                      'values'),
                values=sorted(range_types),
                ),
        Bool('bidirectional?',
             label=_('Two-way trust'),
             cli_name='two_way',
             doc=(_('Establish bi-directional trust. By default trust is '
                    'inbound one-way only.')),
             default=False,
             ),
        Bool('external?',
             label=_('External trust'),
             cli_name='external',
             doc=_('Establish external trust to a domain in another forest. '
                   'The trust is not transitive beyond the domain.'),
             default=False,
             ),
    )

    msg_summary = _('Added Active Directory trust for realm "%(value)s"')
    msg_summary_existing = _('Re-established trust to domain "%(value)s"')

    def _format_trust_attrs(self, result, **options):

        # Format the output into human-readable values
        attributes = int(result['result'].get('ipanttrustattributes', [0])[0])

        if not options.get('raw', False):
            result['result']['trusttype'] = [trust_type_string(
                result['result']['ipanttrusttype'][0], attributes)]
            result['result']['trustdirection'] = [trust_direction_string(
                result['result']['ipanttrustdirection'][0])]
            result['result']['truststatus'] = [trust_status_string(
                result['verified'])]

        if attributes:
            result['result'].pop('ipanttrustattributes', None)

        result['result'].pop('ipanttrustauthoutgoing', None)
        result['result'].pop('ipanttrustauthincoming', None)

    def execute(self, *keys, **options):
        ldap = self.obj.backend

        verify_samba_component_presence(ldap, self.api)

        full_join = self.validate_options(*keys, **options)
        old_range, range_name, dom_sid = self.validate_range(*keys, **options)
        result = self.execute_ad(full_join, *keys, **options)

        if not old_range:
            # Store the created range type, since for POSIX trusts no
            # ranges for the subdomains should be added, POSIX attributes
            # provide a global mapping across all subdomains
            add_range(
                self.api, self.trustinstance, range_name, dom_sid,
                *keys, **options
            )

        attrs_list = self.obj.default_attributes
        if options.get('all', False):
            attrs_list.append('*')

        trust_filter = "cn=%s" % result['value']
        trusts, _truncated = ldap.find_entries(
            base_dn=DN(self.api.env.container_trusts, self.api.env.basedn),
            filter=trust_filter,
            attrs_list=attrs_list
        )

        result['result'] = entry_to_dict(trusts[0], **options)

        # Fetch topology of the trust forest -- we need always to do it
        # for AD trusts, regardless of the type of idranges associated with it
        # Note that add_new_domains_from_trust will add needed ranges for
        # the algorithmic ID mapping case.
        if (options.get('trust_type') == u'ad' and
                options.get('trust_secret') is None):

            if options.get('bidirectional') == True:
                # Bidirectional trust allows us to use cross-realm TGT,
                # so we can run the call under original user's credentials
                res = fetch_domains_from_trust(self.api, self.trustinstance,
                                               **options)
                add_new_domains_from_trust(
                    self.api, self.trustinstance, result['result'], res,
                    **options)
            else:
                # One-way trust is more complex. We don't have cross-realm TGT
                # and cannot use IPA principals to authenticate against AD.
                # Instead, we have to use our trusted domain object's (TDO)
                # account in AD. Access to the credentials is limited and IPA
                # framework cannot access it directly.  Instead, we call out to
                # oddjobd-activated higher privilege process that will use TDO
                # object credentials to authenticate to AD with Kerberos,
                # run DCE RPC calls to do discovery and will call
                # add_new_domains_from_trust() on its own.
                fetch_trusted_domains_over_dbus(self.api, result['value'])

        # Format the output into human-readable values unless `--raw` is given
        self._format_trust_attrs(result, **options)
        del result['verified']

        return result

    def validate_options(self, *keys, **options):
        trusted_realm_domain = keys[-1]

        if not _murmur_installed and 'base_id' not in options:
            raise errors.ValidationError(
                name=_('missing base_id'),
                error=_(
                    'pysss_murmur is not available on the server '
                    'and no base-id is given.'
                )
            )

        if 'trust_type' not in options:
            raise errors.RequirementError(name='trust_type')

        if options['trust_type'] != u'ad':
            raise errors.ValidationError(
                name=_('trust type'),
                error=_('only "ad" is supported')
            )

        # Detect IPA-AD domain clash
        if self.api.env.domain.lower() == trusted_realm_domain.lower():
            raise errors.ValidationError(
                name=_('domain'),
                error=_('Cannot establish a trust to AD deployed in the same '
                        'domain as IPA. Such setup is not supported.')
                )

        # If domain name and realm does not match, IPA server is not be able
        # to establish trust with Active Directory.

        realm_not_matching_domain = (
            self.api.env.domain.upper() != self.api.env.realm
        )

        if options['trust_type'] == u'ad' and realm_not_matching_domain:
            raise errors.ValidationError(
                name=_('Realm-domain mismatch'),
                error=_('To establish trust with Active Directory, the '
                        'domain name and the realm name of the IPA server '
                        'must match')
                )

        self.trustinstance = ipaserver.dcerpc.TrustDomainJoins(self.api)
        if not self.trustinstance.configured:
            raise errors.NotFound(
                name=_('AD Trust setup'),
                reason=_(
                    'Cannot perform join operation without own domain '
                    'configured. Make sure you have run ipa-adtrust-install '
                    'on the IPA server first'
                )
            )

        # Obtain a list of IPA realm domains
        result = self.api.Command.realmdomains_show()['result']
        realm_domains = result['associateddomain']

        # Do not allow the AD's trusted realm domain in the list
        # of our realm domains
        if trusted_realm_domain.lower() in realm_domains:
            raise errors.ValidationError(
                name=_('AD Trust setup'),
                error=_(
                    'Trusted domain %(domain)s is included among '
                    'IPA realm domains. It needs to be removed '
                    'prior to establishing the trust. See the '
                    '"ipa realmdomains-mod --del-domain" command.'
                ) % dict(domain=trusted_realm_domain)
            )

        self.realm_server = options.get('realm_server')
        self.realm_admin = options.get('realm_admin')
        self.realm_passwd = options.get('realm_passwd')

        if self.realm_admin:
            names = self.realm_admin.split('@')

            if len(names) > 1:
                # realm admin name is in UPN format, user@realm, check that
                # realm is the same as the one that we are attempting to trust
                if trusted_realm_domain.lower() != names[-1].lower():
                    raise errors.ValidationError(
                        name=_('AD Trust setup'),
                        error=_(
                            'Trusted domain and administrator account use '
                            'different realms'
                        )
                    )
                self.realm_admin = names[0]

            if not self.realm_passwd:
                raise errors.ValidationError(
                    name=_('AD Trust setup'),
                    error=_('Realm administrator password should be specified')
                )
            return True

        return False

    def validate_range(self, *keys, **options):
        # If a range for this trusted domain already exists,
        # '--base-id' or '--range-size' options should not be specified
        range_name = keys[-1].upper() + '_id_range'
        range_type = options.get('range_type')

        try:
            old_range = self.api.Command['idrange_show'](range_name, raw=True)
        except errors.NotFound:
            old_range = None

        if options.get('trust_type') == u'ad':
            if range_type and range_type not in (u'ipa-ad-trust',
                                                 u'ipa-ad-trust-posix'):
                raise errors.ValidationError(
                    name=_('id range type'),
                    error=_(
                        'Only the ipa-ad-trust and ipa-ad-trust-posix are '
                        'allowed values for --range-type when adding an AD '
                        'trust.'
                    ))

        base_id = options.get('base_id')
        range_size = options.get('range_size')

        if old_range and (base_id or range_size):
            raise errors.ValidationError(
                name=_('id range'),
                error=_(
                    'An id range already exists for this trust. '
                    'You should either delete the old range, or '
                    'exclude --base-id/--range-size options from the command.'
                )
            )

        # If a range for this trusted domain already exists,
        # domain SID must also match
        self.trustinstance.populate_remote_domain(
            keys[-1],
            self.realm_server,
            self.realm_admin,
            self.realm_passwd
        )
        dom_sid = self.trustinstance.remote_domain.info['sid']

        if old_range:
            old_dom_sid = old_range['result']['ipanttrusteddomainsid'][0]
            old_range_type = old_range['result']['iparangetype'][0]

            if old_dom_sid != dom_sid:
                raise errors.ValidationError(
                    name=_('range exists'),
                    error=_(
                        'ID range with the same name but different domain SID '
                        'already exists. The ID range for the new trusted '
                        'domain must be created manually.'
                    )
                )

            if range_type and range_type != old_range_type:
                raise errors.ValidationError(
                    name=_('range type change'),
                    error=_('ID range for the trusted domain already '
                            'exists, but it has a different type. Please '
                            'remove the old range manually, or do not '
                            'enforce type via --range-type option.'))

        return old_range, range_name, dom_sid

    def execute_ad(self, full_join, *keys, **options):
        # Join domain using full credentials and with random trustdom
        # secret (will be generated by the join method)

        # First see if the trust is already in place
        # Force retrieval of the trust object by not passing trust_type
        try:
            dn = self.obj.get_dn(keys[-1])
        except errors.NotFound:
            dn = None

        trust_type = TRUST_ONEWAY
        if options.get('bidirectional', False):
            trust_type = TRUST_BIDIRECTIONAL

        # If we are forced to establish external trust, allow it
        if options.get('external', False):
            self.trustinstance.allow_behavior(TRUST_JOIN_EXTERNAL)

        # 1. Full access to the remote domain. Use admin credentials and
        # generate random trustdom password to do work on both sides
        if full_join:
            try:
                result = self.trustinstance.join_ad_full_credentials(
                    keys[-1],
                    self.realm_server,
                    self.realm_admin,
                    self.realm_passwd,
                    trust_type
                )
            except errors.NotFound:
                error_message = _("Unable to resolve domain controller for "
                                  "{domain} domain. "
                                  ).format(domain=keys[-1])
                instructions = []

                if dns_container_exists(self.obj.backend):
                    try:
                        dns_zone = self.api.Command.dnszone_show(
                            keys[-1])['result']

                        if (('idnsforwardpolicy' in dns_zone) and
                                dns_zone['idnsforwardpolicy'][0] == u'only'):

                            instructions.append(
                                _("Forward policy is defined for it in "
                                  "IPA DNS, perhaps forwarder points to "
                                  "incorrect host?")
                            )
                    except (errors.NotFound, KeyError):
                        _instruction = _(
                            "IPA manages DNS, please verify your DNS "
                            "configuration and make sure that service "
                            "records of the '{domain}' domain can be "
                            "resolved. Examples how to configure DNS "
                            "with CLI commands or the Web UI can be "
                            "found in the documentation. "
                        )
                        instructions.append(
                            _instruction.format(domain=keys[-1])
                        )
                else:
                    _instruction = _(
                        "Since IPA does not manage DNS records, ensure "
                        "DNS is configured to resolve '{domain}' "
                        "domain from IPA hosts and back."
                    )
                    instructions.append(
                        _instruction.format(domain=keys[-1])
                    )
                raise errors.NotFound(
                    reason=error_message,
                    instructions=instructions
                )

            if result is None:
                raise errors.ValidationError(
                    name=_('AD Trust setup'),
                    error=_('Unable to verify write permissions to the AD')
                )

            ret = dict(
                value=pkey_to_value(
                    self.trustinstance.remote_domain.info['dns_domain'],
                    options),
                verified=result['verified']
            )
            if dn:
                ret['summary'] = self.msg_summary_existing % ret
            return ret

        # 2. We don't have access to the remote domain and trustdom password
        # is provided. Do the work on our side and inform what to do on remote
        # side.
        if options.get('trust_secret'):
            result = self.trustinstance.join_ad_ipa_half(
                keys[-1],
                self.realm_server,
                options['trust_secret'],
                trust_type
            )
            ret = dict(
                value=pkey_to_value(
                    self.trustinstance.remote_domain.info['dns_domain'],
                    options),
                verified=result['verified']
            )
            if dn:
                ret['summary'] = self.msg_summary_existing % ret
            return ret
        else:
            raise errors.ValidationError(
                name=_('AD Trust setup'),
                error=_('Not enough arguments specified to perform trust '
                        'setup'))


@register()
class trust_del(LDAPDelete):
    __doc__ = _('Delete a trust.')

    msg_summary = _('Deleted trust "%(value)s"')


@register()
class trust_mod(LDAPUpdate):
    __doc__ = _("""
    Modify a trust (for future use).

    Currently only the default option to modify the LDAP attributes is
    available. More specific options will be added in coming releases.
    """)

    msg_summary = _('Modified trust "%(value)s" '
                    '(change will be effective in 60 seconds)')

    def pre_callback(self, ldap, dn, e_attrs, attrs_list, *keys, **options):
        assert isinstance(dn, DN)

        self.obj.validate_sid_blacklists(e_attrs)

        return dn


@register()
class trust_find(LDAPSearch):
    __doc__ = _('Search for trusts.')
    has_output_params = (LDAPSearch.has_output_params +
                         (Str('ipanttrusttype'), Str('ipanttrustattributes')))

    msg_summary = ngettext(
        '%(count)d trust matched', '%(count)d trusts matched', 0
    )

    # Since all trusts types are stored within separate containers
    # under 'cn=trusts', search needs to be done on a sub-tree scope
    def pre_callback(self, ldap, filters, attrs_list,
                     base_dn, scope, *args, **options):
        # list only trust, not trust domains
        return (filters, base_dn, ldap.SCOPE_SUBTREE)

    def execute(self, *args, **options):
        result = super(trust_find, self).execute(*args, **options)

        self.obj.warning_if_ad_trust_dom_have_missing_SID(result, **options)

        return result

    def post_callback(self, ldap, entries, truncated, *args, **options):
        if options.get('pkey_only', False):
            return truncated

        for attrs in entries:
            # Translate ipanttrusttype to trusttype if --raw not used
            trust_type = attrs.single_value.get('ipanttrusttype', None)
            attributes = attrs.single_value.get('ipanttrustattributes', 0)
            if not options.get('raw', False) and trust_type is not None:
                attrs['trusttype'] = [
                    trust_type_string(trust_type, attributes)
                ]
                del attrs['ipanttrusttype']
                if attributes:
                    del attrs['ipanttrustattributes']

        return truncated


@register()
class trust_show(LDAPRetrieve):
    __doc__ = _('Display information about a trust.')
    has_output_params = (LDAPRetrieve.has_output_params +
                         (Str('ipanttrusttype'),
                          Str('ipanttrustdirection'),
                          Str('ipanttrustattributes')))

    def execute(self, *keys, **options):
        result = super(trust_show, self).execute(*keys, **options)

        self.obj.warning_if_ad_trust_dom_have_missing_SID(result, **options)

        return result

    def post_callback(self, ldap, dn, e_attrs, *keys, **options):

        assert isinstance(dn, DN)
        # Translate ipanttrusttype to trusttype
        # and ipanttrustdirection to trustdirection
        # if --raw not used

        if not options.get('raw', False):
            trust_type = e_attrs.single_value.get('ipanttrusttype', None)
            attributes = e_attrs.single_value.get('ipanttrustattributes', 0)
            if trust_type is not None:
                e_attrs['trusttype'] = [
                    trust_type_string(trust_type, attributes)
                ]
                del e_attrs['ipanttrusttype']

            dir_str = e_attrs.single_value.get('ipanttrustdirection', None)
            if dir_str is not None:
                e_attrs['trustdirection'] = [trust_direction_string(dir_str)]
                del e_attrs['ipanttrustdirection']

            if attributes:
                del e_attrs['ipanttrustattributes']

        return dn


_trustconfig_dn = {
    u'ad': DN(('cn', api.env.domain),
              api.env.container_cifsdomains, api.env.basedn),
}


@register()
class trustconfig(LDAPObject):
    """
    Trusts global configuration object
    """
    object_name = _('trust configuration')
    default_attributes = [
        'cn', 'ipantsecurityidentifier', 'ipantflatname', 'ipantdomainguid',
        'ipantfallbackprimarygroup',
    ]

    label = _('Global Trust Configuration')
    label_singular = _('Global Trust Configuration')

    takes_params = (
        Str('cn',
            label=_('Domain'),
            flags=['no_update'],
        ),
        Str('ipantsecurityidentifier',
            label=_('Security Identifier'),
            flags=['no_update'],
        ),
        Str('ipantflatname',
            label=_('NetBIOS name'),
            flags=['no_update'],
        ),
        Str('ipantdomainguid',
            label=_('Domain GUID'),
            flags=['no_update'],
        ),
        Str('ipantfallbackprimarygroup',
            cli_name='fallback_primary_group',
            label=_('Fallback primary group'),
        ),
        Str(
            'ad_trust_agent_server*',
            label=_('IPA AD trust agents'),
            doc=_('IPA servers configured as AD trust agents'),
            flags={'virtual_attribute', 'no_create', 'no_update'}
        ),
        Str(
            'ad_trust_controller_server*',
            label=_('IPA AD trust controllers'),
            doc=_('IPA servers configured as AD trust controllers'),
            flags={'virtual_attribute', 'no_create', 'no_update'}
        ),
    )

    def get_dn(self, *keys, **kwargs):
        trust_type = kwargs.get('trust_type')
        if trust_type is None:
            raise errors.RequirementError(name='trust_type')
        try:
            return _trustconfig_dn[kwargs['trust_type']]
        except KeyError:
            raise errors.ValidationError(
                name='trust_type',
                error=_("unsupported trust type")
            )

    def _normalize_groupdn(self, entry_attrs):
        """
        Checks that group with given name/DN exists and updates the entry_attrs
        """
        if 'ipantfallbackprimarygroup' not in entry_attrs:
            return

        group = entry_attrs['ipantfallbackprimarygroup']
        if isinstance(group, (list, tuple)):
            group = group[0]

        if group is None:
            return

        try:
            dn = DN(group)
            # group is in a form of a DN
            try:
                self.backend.get_entry(dn)
            except errors.NotFound:
                raise self.api.Object['group'].handle_not_found(group)
            # DN is valid, we can just return
            return
        except ValueError:
            # The search is performed for groups with "posixgroup" objectclass
            # and not "ipausergroup" so that it can also match groups like
            # "Default SMB Group" which does not have this objectclass.
            try:
                group_entry = self.backend.find_entry_by_attr(
                    self.api.Object['group'].primary_key.name,
                    group,
                    ['posixgroup'],
                    [''],
                    DN(self.api.env.container_group, self.api.env.basedn))
            except errors.NotFound:
                raise self.api.Object['group'].handle_not_found(group)
            else:
                entry_attrs['ipantfallbackprimarygroup'] = [group_entry.dn]

    def _convert_groupdn(self, entry_attrs, options):
        """
        Convert an group dn into a name. As we use CN as user RDN, its value
        can be extracted from the DN without further LDAP queries.
        """
        if options.get('raw', False):
            return

        try:
            groupdn = entry_attrs['ipantfallbackprimarygroup'][0]
        except (IndexError, KeyError):
            groupdn = None

        if groupdn is None:
            return
        assert isinstance(groupdn, DN)

        entry_attrs['ipantfallbackprimarygroup'] = [groupdn[0][0].value]


@register()
class trustconfig_mod(LDAPUpdate):
    __doc__ = _('Modify global trust configuration.')

    takes_options = LDAPUpdate.takes_options + (_trust_type_option,)
    msg_summary = _('Modified "%(value)s" trust configuration')
    has_output = output.simple_entry

    def pre_callback(self, ldap, dn, e_attrs, attrs_list, *keys, **options):
        self.obj._normalize_groupdn(e_attrs)
        return dn

    def execute(self, *keys, **options):
        result = super(trustconfig_mod, self).execute(*keys, **options)
        result['value'] = pkey_to_value(options['trust_type'], options)
        return result

    def post_callback(self, ldap, dn, e_attrs, *keys, **options):
        self.obj._convert_groupdn(e_attrs, options)
        self.api.Object.config.show_servroles_attributes(
            e_attrs, "AD trust agent", "AD trust controller", **options)
        return dn


@register()
class trustconfig_show(LDAPRetrieve):
    __doc__ = _('Show global trust configuration.')

    takes_options = LDAPRetrieve.takes_options + (_trust_type_option,)
    has_output = output.simple_entry

    def execute(self, *keys, **options):
        result = super(trustconfig_show, self).execute(*keys, **options)
        result['value'] = pkey_to_value(options['trust_type'], options)
        return result

    def post_callback(self, ldap, dn, entry_attrs, *keys, **options):
        self.obj._convert_groupdn(entry_attrs, options)
        self.api.Object.config.show_servroles_attributes(
            entry_attrs, "AD trust agent", "AD trust controller", **options)

        return dn


if _nss_idmap_installed:
    _idmap_type_dict = {
        pysss_nss_idmap.ID_USER: 'user',
        pysss_nss_idmap.ID_GROUP: 'group',
        pysss_nss_idmap.ID_BOTH: 'both',
    }

    def idmap_type_string(level):
        string = _idmap_type_dict.get(int(level), 'unknown')
        return unicode(string)


@register()
class trust_resolve(Command):
    NO_CLI = True
    __doc__ = _('Resolve security identifiers of users and groups '
                'in trusted domains')

    takes_options = (
        Str('sids+',
            label = _('Security Identifiers (SIDs)'),
        ),
    )

    has_output_params = (
        Str('name', label=_('Name')),
        Str('sid', label=_('SID')),
    )

    has_output = (
        output.ListOfEntries('result'),
    )

    def execute(self, *keys, **options):
        result = list()
        if not _nss_idmap_installed:
            return dict(result=result)
        try:
            NAME_KEY = pysss_nss_idmap.NAME_KEY
            TYPE_KEY = pysss_nss_idmap.TYPE_KEY
            sids = [str(x) for x in options['sids']]
            xlate = pysss_nss_idmap.getnamebysid(sids)
            for sid in xlate:
                entry = dict()
                entry['sid'] = [unicode(sid)]
                entry['name'] = [unicode(xlate[sid][NAME_KEY])]
                entry['type'] = [idmap_type_string(xlate[sid][TYPE_KEY])]
                result.append(entry)
        except ValueError:
            pass

        return dict(result=result)


@register()
class adtrust_is_enabled(Command):
    NO_CLI = True

    __doc__ = _('Determine whether ipa-adtrust-install has been run on this '
                'system')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        adtrust_dn = DN(
            ('cn', 'ADTRUST'),
            ('cn', self.api.env.host),
            ('cn', 'masters'),
            ('cn', 'ipa'),
            ('cn', 'etc'),
            self.api.env.basedn
        )

        try:
            ldap.get_entry(adtrust_dn)
        except errors.NotFound:
            return dict(result=False)

        return dict(result=True)


@register()
class compat_is_enabled(Command):
    NO_CLI = True

    __doc__ = _('Determine whether Schema Compatibility plugin is configured '
                'to serve trusted domain users and groups')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        users_dn = DN(
            ('cn', 'users'),
            ('cn', 'Schema Compatibility'),
            ('cn', 'plugins'),
            ('cn', 'config')
        )
        groups_dn = DN(
            ('cn', 'groups'),
            ('cn', 'Schema Compatibility'),
            ('cn', 'plugins'),
            ('cn', 'config')
        )

        try:
            users_entry = ldap.get_entry(users_dn)
        except errors.NotFound:
            return dict(result=False)

        attr = users_entry.get('schema-compat-lookup-nsswitch')
        if not attr or 'user' not in attr:
            return dict(result=False)

        try:
            groups_entry = ldap.get_entry(groups_dn)
        except errors.NotFound:
            return dict(result=False)

        attr = groups_entry.get('schema-compat-lookup-nsswitch')
        if not attr or 'group' not in attr:
            return dict(result=False)

        return dict(result=True)


@register()
class sidgen_was_run(Command):
    """
    This command tries to determine whether the sidgen task was run during
    ipa-adtrust-install. It does that by simply checking the "editors" group
    for the presence of the ipaNTSecurityIdentifier attribute - if the
    attribute is present, the sidgen task was run.

    Since this command relies on the existence of the "editors" group, it will
    fail loudly in case this group does not exist.
    """
    NO_CLI = True

    __doc__ = _('Determine whether ipa-adtrust-install has been run with '
                'sidgen task')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        editors_dn = DN(
            ('cn', 'editors'),
            ('cn', 'groups'),
            ('cn', 'accounts'),
            api.env.basedn
        )

        try:
            editors_entry = ldap.get_entry(editors_dn)
        except errors.NotFound:
            raise errors.NotFound(
                name=_('sidgen_was_run'),
                reason=_(
                    'This command relies on the existence of the "editors" '
                    'group, but this group was not found.'
                )
            )

        attr = editors_entry.get('ipaNTSecurityIdentifier')
        if not attr:
            return dict(result=False)

        return dict(result=True)


@register()
class trustdomain(LDAPObject):
    """
    Object representing a domain of the AD trust.
    """
    parent_object = 'trust'
    trust_type_idx = {'2': u'ad'}
    object_name = _('trust domain')
    object_name_plural = _('trust domains')
    object_class = ['ipaNTTrustedDomain']
    default_attributes = ['cn', 'ipantflatname', 'ipanttrusteddomainsid',
                          'ipanttrustpartner', 'ipantadditionalsuffixes']
    search_display_attributes = ['cn', 'ipantflatname',
                                 'ipanttrusteddomainsid',
                                 'ipantadditionalsuffixes']

    label = _('Trusted domains')
    label_singular = _('Trusted domain')

    takes_params = (
        Str('cn',
            label=_('Domain name'),
            cli_name='domain',
            primary_key=True),
        Str('ipantflatname?',
            cli_name='flat_name',
            label=_('Domain NetBIOS name')),
        Str('ipanttrusteddomainsid?',
            cli_name='sid',
            label=_('Domain Security Identifier')),
        Flag('domain_enabled',
             label=_('Domain enabled'),
             flags={'virtual_attribute',
                    'no_create', 'no_update', 'no_search'}),
    )

    # LDAPObject.get_dn() only passes all but last element of keys and no
    # kwargs to the parent object's get_dn() no matter what you pass to it.
    # Make own get_dn() as we really need all elements to construct proper dn.
    def get_dn(self, *keys, **kwargs):
        sdn = [('cn', x) for x in keys]
        sdn.reverse()
        trust_type = kwargs.get('trust_type')
        if not trust_type:
            trust_type = u'ad'

        dn = make_trust_dn(self.env, trust_type, DN(*sdn))
        return dn


@register()
class trustdomain_find(LDAPSearch):
    __doc__ = _('Search domains of the trust')

    def pre_callback(self, ldap, filters, attrs_list, base_dn,
                     scope, *args, **options):
        return (filters, base_dn, ldap.SCOPE_SUBTREE)

    def post_callback(self, ldap, entries, truncated, *args, **options):
        if options.get('pkey_only', False):
            return truncated
        trust_dn = self.obj.get_dn(args[0], trust_type=u'ad')
        trust_entry = ldap.get_entry(trust_dn)
        blacklist = trust_entry.get('ipantsidblacklistincoming')
        for entry in entries:
            sid = entry.get('ipanttrusteddomainsid', [None])[0]
            if sid is None:
                continue

            if sid in blacklist:
                entry['domain_enabled'] = [False]
            else:
                entry['domain_enabled'] = [True]
        return truncated


@register()
class trustdomain_mod(LDAPUpdate):
    __doc__ = _('Modify trustdomain of the trust')

    NO_CLI = True
    takes_options = LDAPUpdate.takes_options + (_trust_type_option,)


@register()
class trustdomain_add(LDAPCreate):
    __doc__ = _('Allow access from the trusted domain')
    NO_CLI = True

    takes_options = LDAPCreate.takes_options + (_trust_type_option,)

    def pre_callback(self, ldap, dn, e_attrs, attrs_list, *keys, **options):
        # ipaNTTrustPartner must always be set to the name of the trusted
        # domain. See MS-ADTS 6.1.6.7.13
        e_attrs['ipanttrustpartner'] = [dn[0]['cn']]
        return dn


@register()
class trustdomain_del(LDAPDelete):
    __doc__ = _('Remove information about the domain associated '
                'with the trust.')

    msg_summary = _('Removed information about the trusted domain '
                    '"%(value)s"')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        verify_samba_component_presence(ldap, self.api)

        # Note that pre-/post- callback handling for LDAPDelete is causing
        # pre_callback to always receive empty keys. We need to catch the case
        # when root domain is being deleted

        for domain in keys[1]:
            try:
                self.obj.get_dn_if_exists(keys[0], domain, trust_type=u'ad')
            except errors.NotFound:
                if keys[0].lower() == domain:
                    raise errors.ValidationError(
                        name='domain',
                        error=_("cannot delete root domain of the trust, "
                                "use trust-del to delete the trust itself"))
                raise self.obj.handle_not_found(keys[0], domain)

            try:
                self.api.Command.trustdomain_enable(keys[0], domain)
            except errors.AlreadyActive:
                pass

        result = super(trustdomain_del, self).execute(*keys, **options)
        result['value'] = pkey_to_value(keys[1], options)
        return result


def fetch_domains_from_trust(myapi, trustinstance, **options):
    """
    Contact trust forest root DC and fetch trusted forest topology information.

    :param myapi: API instance
    :param trustinstance: Initialized instance of `dcerpc.TrustDomainJoins`
        class
    :param options: options passed from API command's `execute()` method

    :returns: dict containing forest domain information and forest-wide UPN
        suffixes (if any)
    """

    forest_root_name = trustinstance.remote_domain.info['dns_forest']

    # We want to use Kerberos if we have admin credentials even with SMB calls
    # as eventually use of NTLMSSP will be deprecated for trusted domain
    # operations If admin credentials are missing, 'creds' will be None and
    # fetch_domains will use HTTP/ipa.master@IPA.REALM principal, e.g. Kerberos
    # authentication as well.
    creds = generate_creds(trustinstance, style=CRED_STYLE_KERBEROS, **options)
    server = options.get('realm_server', None)
    domains = ipaserver.dcerpc.fetch_domains(
        myapi, trustinstance.local_flatname, forest_root_name, creds=creds,
        server=server)

    return domains


def add_new_domains_from_trust(myapi, trustinstance, trust_entry,
                               domains, **options):
    result = []
    if not domains:
        return result

    trust_name = trust_entry['cn'][0]
    # trust range must exist by the time add_new_domains_from_trust is called
    range_name = trust_name.upper() + '_id_range'
    old_range = myapi.Command.idrange_show(range_name, raw=True)['result']
    idrange_type = old_range['iparangetype'][0]

    suffixes = list()
    suffixes.extend(y['cn']
                    for x, y in six.iteritems(domains['suffixes'])
                    if x not in domains['domains'])

    try:
        dn = myapi.Object.trust.get_dn(trust_name, trust_type=u'ad')
        ldap = myapi.Backend.ldap2
        entry = ldap.get_entry(dn)
        tlns = entry.get('ipantadditionalsuffixes', [])
        tlns.extend(x for x in suffixes if x not in tlns)
        entry['ipantadditionalsuffixes'] = tlns
        ldap.update_entry(entry)
    except errors.EmptyModlist:
        pass

    is_nontransitive = int(trust_entry.get('ipanttrustattributes',
                           [0])[0]) & LSA_TRUST_ATTRIBUTE_NON_TRANSITIVE

    if is_nontransitive:
        return result

    for dom in six.itervalues(domains['domains']):
        dom['trust_type'] = u'ad'
        try:
            name = dom['cn']
            del dom['cn']
            if 'all' in options:
                dom['all'] = options['all']
            if 'raw' in options:
                dom['raw'] = options['raw']

            try:
                res = myapi.Command.trustdomain_add(trust_name, name, **dom)
                result.append(res['result'])
            except errors.DuplicateEntry:
                # Ignore updating duplicate entries
                pass

            if idrange_type != u'ipa-ad-trust-posix':
                range_name = name.upper() + '_id_range'
                dom['range_type'] = u'ipa-ad-trust'
                add_range(myapi, trustinstance,
                          range_name, dom['ipanttrusteddomainsid'],
                          name, **dom)
        except errors.DuplicateEntry:
            # Ignore updating duplicate entries
            pass

    return result


@register()
class trust_fetch_domains(LDAPRetrieve):
    __doc__ = _('Refresh list of the domains associated with the trust')

    has_output = output.standard_list_of_entries
    takes_options = LDAPRetrieve.takes_options + (
        Str('realm_admin?',
            cli_name='admin',
            label=_("Active Directory domain administrator"),
            ),
        Password('realm_passwd?',
                 cli_name='password',
                 label=_("Active Directory domain administrator's password"),
                 confirm=False,
                 ),
        Str('realm_server?',
            cli_name='server',
            label=_('Domain controller for the Active Directory domain '
                    '(optional)'),
            ),
    )

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        verify_samba_component_presence(ldap, self.api)

        # Check first that the trust actually exists
        result = self.api.Command.trust_show(keys[0], all=True, raw=True)
        self.obj.warning_if_ad_trust_dom_have_missing_SID(result, **options)

        result = dict()
        result['result'] = []
        result['count'] = 0
        result['truncated'] = False

        # For one-way trust and external trust fetch over DBus.
        # We don't get the list in this case.
        # With privilege separation we also cannot authenticate as
        # HTTP/ principal because we have no access to its key material.
        # Thus, we'll use DBus call out to oddjobd helper in all cases
        fetch_trusted_domains_over_dbus(self.api, *keys, **options)
        result['summary'] = unicode(_('List of trust domains successfully '
                                      'refreshed. Use trustdomain-find '
                                      'command to list them.'))
        return result


@register()
class trustdomain_enable(LDAPQuery):
    __doc__ = _('Allow use of IPA resources by the domain of the trust')

    has_output = output.standard_value
    msg_summary = _('Enabled trust domain "%(value)s"')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        verify_samba_component_presence(ldap, self.api)

        if keys[0].lower() == keys[1].lower():
            raise errors.ValidationError(
                name='domain',
                error=_("Root domain of the trust is always enabled "
                        "for the existing trust")
            )
        try:
            trust_dn = self.obj.get_dn(keys[0], trust_type=u'ad')
            trust_entry = ldap.get_entry(trust_dn)
        except errors.NotFound:
            raise self.api.Object[self.obj.parent_object].handle_not_found(
                keys[0])

        dn = self.obj.get_dn(keys[0], keys[1], trust_type=u'ad')
        try:
            entry = ldap.get_entry(dn)
            sid = entry.single_value.get('ipanttrusteddomainsid', None)
            if sid in trust_entry['ipantsidblacklistincoming']:
                trust_entry['ipantsidblacklistincoming'].remove(sid)
                ldap.update_entry(trust_entry)
            else:
                raise errors.AlreadyActive()
        except errors.NotFound:
            raise self.obj.handle_not_found(*keys)

        return dict(
            result=True,
            value=pkey_to_value(keys[1], options),
        )


@register()
class trustdomain_disable(LDAPQuery):
    __doc__ = _('Disable use of IPA resources by the domain of the trust')

    has_output = output.standard_value
    msg_summary = _('Disabled trust domain "%(value)s"')

    def execute(self, *keys, **options):
        ldap = self.api.Backend.ldap2
        verify_samba_component_presence(ldap, self.api)

        if keys[0].lower() == keys[1].lower():
            raise errors.ValidationError(
                name='domain',
                error=_("cannot disable root domain of the trust, "
                        "use trust-del to delete the trust itself")
            )
        try:
            trust_dn = self.obj.get_dn(keys[0], trust_type=u'ad')
            trust_entry = ldap.get_entry(trust_dn)
        except errors.NotFound:
            raise self.api.Object[self.obj.parent_object].handle_not_found(
                keys[0])

        dn = self.obj.get_dn(keys[0], keys[1], trust_type=u'ad')
        try:
            entry = ldap.get_entry(dn)
            sid = entry.single_value.get('ipanttrusteddomainsid', None)
            if sid not in trust_entry['ipantsidblacklistincoming']:
                trust_entry['ipantsidblacklistincoming'].append(sid)
                ldap.update_entry(trust_entry)
            else:
                raise errors.AlreadyInactive()
        except errors.NotFound:
            raise self.obj.handle_not_found(*keys)

        return dict(
            result=True,
            value=pkey_to_value(keys[1], options),
        )



#
# Copyright (C) 2014  FreeIPA Contributors see COPYING for license
#

import logging

from ipaserver import p11helper as _ipap11helper

logger = logging.getLogger(__name__)

attrs_id2name = {
    #_ipap11helper.CKA_ALLOWED_MECHANISMS: 'ipk11allowedmechanisms',
    _ipap11helper.CKA_ALWAYS_AUTHENTICATE: 'ipk11alwaysauthenticate',
    _ipap11helper.CKA_ALWAYS_SENSITIVE: 'ipk11alwayssensitive',
    #_ipap11helper.CKA_CHECK_VALUE: 'ipk11checkvalue',
    _ipap11helper.CKA_COPYABLE: 'ipk11copyable',
    _ipap11helper.CKA_DECRYPT: 'ipk11decrypt',
    _ipap11helper.CKA_DERIVE: 'ipk11derive',
    #_ipap11helper.CKA_DESTROYABLE: 'ipk11destroyable',
    _ipap11helper.CKA_ENCRYPT: 'ipk11encrypt',
    #_ipap11helper.CKA_END_DATE: 'ipk11enddate',
    _ipap11helper.CKA_EXTRACTABLE: 'ipk11extractable',
    _ipap11helper.CKA_ID: 'ipk11id',
    #_ipap11helper.CKA_KEY_GEN_MECHANISM: 'ipk11keygenmechanism',
    _ipap11helper.CKA_KEY_TYPE: 'ipk11keytype',
    _ipap11helper.CKA_LABEL: 'ipk11label',
    _ipap11helper.CKA_LOCAL: 'ipk11local',
    _ipap11helper.CKA_MODIFIABLE: 'ipk11modifiable',
    _ipap11helper.CKA_NEVER_EXTRACTABLE: 'ipk11neverextractable',
    _ipap11helper.CKA_PRIVATE: 'ipk11private',
    #_ipap11helper.CKA_PUBLIC_KEY_INFO: 'ipapublickey',
    #_ipap11helper.CKA_PUBLIC_KEY_INFO: 'ipk11publickeyinfo',
    _ipap11helper.CKA_SENSITIVE: 'ipk11sensitive',
    _ipap11helper.CKA_SIGN: 'ipk11sign',
    _ipap11helper.CKA_SIGN_RECOVER: 'ipk11signrecover',
    #_ipap11helper.CKA_START_DATE: 'ipk11startdate',
    #_ipap11helper.CKA_SUBJECT: 'ipk11subject',
    _ipap11helper.CKA_TRUSTED: 'ipk11trusted',
    _ipap11helper.CKA_UNWRAP: 'ipk11unwrap',
    #_ipap11helper.CKA_UNWRAP_TEMPLATE: 'ipk11unwraptemplate',
    _ipap11helper.CKA_VERIFY: 'ipk11verify',
    _ipap11helper.CKA_VERIFY_RECOVER: 'ipk11verifyrecover',
    _ipap11helper.CKA_WRAP: 'ipk11wrap',
    #_ipap11helper.CKA_WRAP_TEMPLATE: 'ipk11wraptemplate',
    _ipap11helper.CKA_WRAP_WITH_TRUSTED: 'ipk11wrapwithtrusted',
}

attrs_name2id = {v: k for k, v in attrs_id2name.items()}

# attribute:
# http://www.freeipa.org/page/V4/PKCS11_in_LDAP/Schema#ipk11KeyType
#
# mapping table:
# http://www.freeipa.org/page/V4/PKCS11_in_LDAP/Schema#CK_MECHANISM_TYPE
keytype_name2id = {
        "rsa": _ipap11helper.KEY_TYPE_RSA,
        "aes": _ipap11helper.KEY_TYPE_AES,
        }

keytype_id2name = {v: k for k, v in keytype_name2id.items()}

wrappingmech_name2id = {
        "rsaPkcs": _ipap11helper.MECH_RSA_PKCS,
        "rsaPkcsOaep": _ipap11helper.MECH_RSA_PKCS_OAEP,
        "aesKeyWrap": _ipap11helper.MECH_AES_KEY_WRAP,
        "aesKeyWrapPad": _ipap11helper.MECH_AES_KEY_WRAP_PAD
        }

wrappingmech_id2name = {v: k for k, v in wrappingmech_name2id.items()}


bool_attr_names = set([
    'ipk11alwaysauthenticate',
    'ipk11alwayssensitive',
    'ipk11copyable',
    'ipk11decrypt',
    'ipk11derive',
    'ipk11encrypt',
    'ipk11extractable',
    'ipk11local',
    'ipk11modifiable',
    'ipk11neverextractable',
    'ipk11private',
    'ipk11sensitive',
    'ipk11sign',
    'ipk11signrecover',
    'ipk11trusted',
    'ipk11unwrap',
    'ipk11verify',
    'ipk11verifyrecover',
    'ipk11wrap',
    'ipk11wrapwithtrusted',
])

modifiable_attrs_id2name = {
    _ipap11helper.CKA_DECRYPT: 'ipk11decrypt',
    _ipap11helper.CKA_DERIVE: 'ipk11derive',
    _ipap11helper.CKA_ENCRYPT: 'ipk11encrypt',
    _ipap11helper.CKA_EXTRACTABLE: 'ipk11extractable',
    _ipap11helper.CKA_ID: 'ipk11id',
    _ipap11helper.CKA_LABEL: 'ipk11label',
    _ipap11helper.CKA_SENSITIVE: 'ipk11sensitive',
    _ipap11helper.CKA_SIGN: 'ipk11sign',
    _ipap11helper.CKA_SIGN_RECOVER: 'ipk11signrecover',
    _ipap11helper.CKA_UNWRAP: 'ipk11unwrap',
    _ipap11helper.CKA_VERIFY: 'ipk11verify',
    _ipap11helper.CKA_VERIFY_RECOVER: 'ipk11verifyrecover',
    _ipap11helper.CKA_WRAP: 'ipk11wrap',
}

modifiable_attrs_name2id = {v: k for k, v in modifiable_attrs_id2name.items()}


def sync_pkcs11_metadata(name, source, target):
    """sync ipk11 metadata from source object to target object"""

    # iterate over list of modifiable PKCS#11 attributes - this prevents us
    # from attempting to set read-only attributes like CKA_LOCAL
    for attr in modifiable_attrs_name2id:
        if attr in source:
            if source[attr] != target[attr]:
                logger.debug('%s: Updating attribute %s from "%s" to "%s"',
                             name,
                             attr,
                             repr(source[attr]),
                             repr(target[attr]))
                target[attr] = source[attr]

def populate_pkcs11_metadata(source, target):
    """populate all ipk11 metadata attributes in target object from source object"""
    for attr in attrs_name2id:
        if attr in source:
            target[attr] = source[attr]

def ldap2p11helper_api_params(ldap_key):
    """prepare dict with metadata parameters suitable for key unwrapping"""
    unwrap_params = {}

    # some attributes are just renamed
    direct_param_map = {
            "ipk11label": "label",
            "ipk11id": "id",
            "ipk11copyable": "cka_copyable",
            "ipk11decrypt": "cka_decrypt",
            "ipk11derive": "cka_derive",
            "ipk11encrypt": "cka_encrypt",
            "ipk11extractable": "cka_extractable",
            "ipk11modifiable": "cka_modifiable",
            "ipk11private": "cka_private",
            "ipk11sensitive": "cka_sensitive",
            "ipk11sign": "cka_sign",
            "ipk11unwrap": "cka_unwrap",
            "ipk11verify": "cka_verify",
            "ipk11wrap": "cka_wrap",
            "ipk11wrapwithtrusted": "cka_wrap_with_trusted"
            }

    for ldap_name, p11h_name in direct_param_map.items():
        if ldap_name in ldap_key:
            unwrap_params[p11h_name] = ldap_key[ldap_name]

    # and some others needs conversion

    indirect_param_map = {
            "ipk11keytype": ("key_type", keytype_name2id),
            "ipawrappingmech": ("wrapping_mech", wrappingmech_name2id),
            }

    for ldap_name, rules in indirect_param_map.items():
        p11h_name, mapping = rules
        if ldap_name in ldap_key:
            unwrap_params[p11h_name] = mapping[ldap_key[ldap_name]]

    return unwrap_params


class AbstractHSM:
    def _filter_replica_keys(self, all_keys):
        replica_keys = {}
        for key_id, key in all_keys.items():
            if not key['ipk11label'].startswith('dnssec-replica:'):
                continue
            replica_keys[key_id] = key
        return replica_keys

    def _filter_zone_keys(self, all_keys):
        zone_keys = {}
        for key_id, key in all_keys.items():
            if key['ipk11label'] == u'dnssec-master' \
                or key['ipk11label'].startswith('dnssec-replica:'):
                continue
            zone_keys[key_id] = key
        return zone_keys



# Authors:
#   Nathaniel McCallum <npmccallum@redhat.com>
#
# Copyright (C) 2014  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipalib import _, api, Int
from ipalib.plugable import Registry
from .baseldap import DN, LDAPObject, LDAPUpdate, LDAPRetrieve

__doc__ = _("""
OTP configuration

Manage the default values that IPA uses for OTP tokens.

EXAMPLES:

 Show basic OTP configuration:
   ipa otpconfig-show

 Show all OTP configuration options:
   ipa otpconfig-show --all

 Change maximum TOTP authentication window to 10 minutes:
   ipa otpconfig-mod --totp-auth-window=600

 Change maximum TOTP synchronization window to 12 hours:
   ipa otpconfig-mod --totp-sync-window=43200

 Change maximum HOTP authentication window to 5:
   ipa hotpconfig-mod --hotp-auth-window=5

 Change maximum HOTP synchronization window to 50:
   ipa hotpconfig-mod --hotp-sync-window=50
""")

register = Registry()

topic = 'otp'


@register()
class otpconfig(LDAPObject):
    object_name = _('OTP configuration options')
    default_attributes = [
        'ipatokentotpauthwindow',
        'ipatokentotpsyncwindow',
        'ipatokenhotpauthwindow',
        'ipatokenhotpsyncwindow',
    ]

    container_dn = DN(('cn', 'otp'), ('cn', 'etc'))
    permission_filter_objectclasses = ['ipatokenotpconfig']
    managed_permissions = {
        'System: Read OTP Configuration': {
            'replaces_global_anonymous_aci': True,
            'ipapermbindruletype': 'all',
            'ipapermright': {'read', 'search', 'compare'},
            'ipapermdefaultattr': {
                'ipatokentotpauthwindow', 'ipatokentotpsyncwindow',
                'ipatokenhotpauthwindow', 'ipatokenhotpsyncwindow',
                'cn',
            },
        },
    }

    label = _('OTP Configuration')
    label_singular = _('OTP Configuration')

    takes_params = (
        Int('ipatokentotpauthwindow',
            cli_name='totp_auth_window',
            label=_('TOTP authentication Window'),
            doc=_('TOTP authentication time variance (seconds)'),
            minvalue=5,
        ),
        Int('ipatokentotpsyncwindow',
            cli_name='totp_sync_window',
            label=_('TOTP Synchronization Window'),
            doc=_('TOTP synchronization time variance (seconds)'),
            minvalue=5,
        ),
        Int('ipatokenhotpauthwindow',
            cli_name='hotp_auth_window',
            label=_('HOTP Authentication Window'),
            doc=_('HOTP authentication skip-ahead'),
            minvalue=1,
        ),
        Int('ipatokenhotpsyncwindow',
            cli_name='hotp_sync_window',
            label=_('HOTP Synchronization Window'),
            doc=_('HOTP synchronization skip-ahead'),
            minvalue=1,
        ),
    )

    def get_dn(self, *keys, **kwargs):
        return self.container_dn + api.env.basedn


@register()
class otpconfig_mod(LDAPUpdate):
    __doc__ = _('Modify OTP configuration options.')


@register()
class otpconfig_show(LDAPRetrieve):
    __doc__ = _('Show the current OTP configuration.')



#
# Copyright (C) 2014  FreeIPA Contributors see COPYING for license
#

import logging

import dns.name
try:
    from xml.etree import cElementTree as etree
except ImportError:
    from xml.etree import ElementTree as etree

from ipapython import ipa_log_manager, ipautil
from ipaplatform.tasks import tasks

logger = logging.getLogger(__name__)

# hack: zone object UUID is stored as path to imaginary zone file
ENTRYUUID_PREFIX = "/var/lib/ipa/dns/zone/entryUUID/"
ENTRYUUID_PREFIX_LEN = len(ENTRYUUID_PREFIX)


class ZoneListReader:
    def __init__(self):
        self.names = set()  # dns.name
        self.uuids = set()  # UUID strings
        self.mapping = dict()      # {UUID: dns.name}

    def _add_zone(self, name, zid):
        """Add zone & UUID to internal structures.

        Zone with given name and UUID must not exist."""
        # detect duplicate zone names
        name = dns.name.from_text(name)
        assert name not in self.names, \
            'duplicate name (%s, %s) vs. %s' % (name, zid, self.mapping)
        # duplicate non-None zid is not allowed
        assert not zid or zid not in self.uuids, \
            'duplicate UUID (%s, %s) vs. %s' % (name, zid, self.mapping)

        self.names.add(name)
        self.uuids.add(zid)
        self.mapping[zid] = name

    def _del_zone(self, name, zid):
        """Remove zone & UUID from internal structures.

        Zone with given name and UUID must exist.
        """
        name = dns.name.from_text(name)
        assert zid is not None
        assert name in self.names, \
            'name (%s, %s) does not exist in %s' % (name, zid, self.mapping)
        assert zid in self.uuids, \
            'UUID (%s, %s) does not exist in %s' % (name, zid, self.mapping)
        assert zid in self.mapping and name == self.mapping[zid], \
            'pair {%s: %s} does not exist in %s' % (zid, name, self.mapping)

        self.names.remove(name)
        self.uuids.remove(zid)
        del self.mapping[zid]


class ODSZoneListReader(ZoneListReader):
    """One-shot parser for ODS zonelist.xml."""
    def __init__(self, zonelist_text):
        super(ODSZoneListReader, self).__init__()
        root = etree.fromstring(zonelist_text)
        self._parse_zonelist(root)

    def _parse_zonelist(self, root):
        """iterate over Zone elements with attribute 'name' and
        add IPA zones to self.zones"""
        if not root.tag == 'ZoneList':
            raise ValueError(root.tag)
        for zone_xml in root.findall('./Zone[@name]'):
            name, zid = self._parse_ipa_zone(zone_xml)
            self._add_zone(name, zid)

    def _parse_ipa_zone(self, zone_xml):
        """Extract zone name, input adapter and detect IPA zones.

        IPA zones have contains Adapters/Input/Adapter element with
        attribute type = "File" and with value prefixed with ENTRYUUID_PREFIX.

        Returns:
            tuple (zone name, ID)
        """
        name = zone_xml.get('name')
        zids = []
        for in_adapter in zone_xml.findall(
                './Adapters/Input/Adapter[@type="File"]'):
            path = in_adapter.text
            if path.startswith(ENTRYUUID_PREFIX):
                # strip prefix from path
                zids.append(path[ENTRYUUID_PREFIX_LEN:])

        if len(zids) != 1:
            raise ValueError('only IPA zones are supported: {}'.format(
                etree.tostring(zone_xml)))

        return name, zids[0]


class LDAPZoneListReader(ZoneListReader):
    def __init__(self):
        super(LDAPZoneListReader, self).__init__()

    def process_ipa_zone(self, op, uuid, zone_ldap):
        assert (op in ['add', 'del']), 'unsupported op %s' % op
        assert uuid is not None
        assert 'idnsname' in zone_ldap, \
            'LDAP zone UUID %s without idnsName' % uuid
        assert len(zone_ldap['idnsname']) == 1, \
            'LDAP zone UUID %s with len(idnsname) != 1' % uuid

        if op == 'add':
            self._add_zone(zone_ldap['idnsname'][0], uuid)
        elif op == 'del':
            self._del_zone(zone_ldap['idnsname'][0], uuid)


class ODSMgr:
    """OpenDNSSEC zone manager. It does LDAP->ODS synchronization.

    Zones with idnsSecInlineSigning attribute = TRUE in LDAP are added
    or deleted from ODS as necessary. ODS->LDAP key synchronization
    has to be solved seperatelly.
    """
    def __init__(self):
        self.zl_ldap = LDAPZoneListReader()

    def ksmutil(self, params):
        """Call ods-ksmutil / ods-enforcer with parameters and return stdout.

        Raises CalledProcessError if returncode != 0.
        """
        result = tasks.run_ods_manager(params, capture_output=True)
        return result.output

    def get_ods_zonelist(self):
        stdout = self.ksmutil(['zonelist', 'export'])
        reader = ODSZoneListReader(stdout)
        return reader

    def add_ods_zone(self, uuid, name):
        zone_path = '%s%s' % (ENTRYUUID_PREFIX, uuid)
        cmd = ['zone', 'add', '--zone', str(name), '--input', zone_path]
        output = self.ksmutil(cmd)
        logger.info('%s', output)
        self.notify_enforcer()

    def del_ods_zone(self, name):
        # ods-ksmutil blows up if zone name has period at the end
        name = name.relativize(dns.name.root)
        # detect if name is root zone
        if name == dns.name.empty:
            name = dns.name.root
        cmd = ['zone', 'delete', '--zone', str(name)]
        output = self.ksmutil(cmd)
        logger.info('%s', output)
        self.notify_enforcer()
        self.cleanup_signer(name)

    def notify_enforcer(self):
        cmd = ['notify']
        output = self.ksmutil(cmd)
        logger.info('%s', output)

    def cleanup_signer(self, zone_name):
        cmd = ['ods-signer', 'ldap-cleanup', str(zone_name)]
        output = ipautil.run(cmd, capture_output=True)
        logger.info('%s', output)

    def ldap_event(self, op, uuid, attrs):
        """Record single LDAP event - zone addition or deletion.

        Change is only recorded to memory.
        self.sync() have to be called to synchronize change to ODS."""
        assert op in ('add', 'del')
        self.zl_ldap.process_ipa_zone(op, uuid, attrs)
        logger.debug("LDAP zones: %s", self.zl_ldap.mapping)

    def sync(self):
        """Synchronize list of zones in LDAP with ODS."""
        zl_ods = self.get_ods_zonelist()
        logger.debug("ODS zones: %s", zl_ods.mapping)
        removed = self.diff_zl(zl_ods, self.zl_ldap)
        logger.info("Zones removed from LDAP: %s", removed)
        added = self.diff_zl(self.zl_ldap, zl_ods)
        logger.info("Zones added to LDAP: %s", added)
        for (uuid, name) in removed:
            self.del_ods_zone(name)
        for (uuid, name) in added:
            self.add_ods_zone(uuid, name)

    def diff_zl(self, s1, s2):
        """Compute zones present in s1 but not present in s2.

        Returns: List of (uuid, name) tuples with zones present only in s1."""
        s1_extra = s1.uuids - s2.uuids
        removed = [(uuid, name) for (uuid, name) in s1.mapping.items()
                   if uuid in s1_extra]
        return removed


if __name__ == '__main__':
    ipa_log_manager.standard_logging_setup(debug=True)
    ods = ODSMgr()
    reader = ods.get_ods_zonelist()
    logger.info('ODS zones: %s', reader.mapping)



#
# Copyright (C) 2016 FreeIPA Contributors see COPYING for license
#

from ipaclient.frontend import MethodOverride
from ipalib import _, errors
from ipalib.plugable import Registry

register = Registry()


@register(override=True, no_fail=True)
class server_del(MethodOverride):
    def interactive_prompt_callback(self, kw):
        server_list = kw.get('cn')
        if not server_list:
            raise errors.RequirementError(name='cn')

        self.api.Backend.textui.print_plain(
            _("Removing %(servers)s from replication topology, "
              "please wait...") % {'servers': ', '.join(server_list)})



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

ENTITY = 'user'

PKEY = 'itest-user'
PASSWD_ITEST_USER = '12345678'
PASSWD_ITEST_USER_NEW = '87654321'

# used for add/delete fixture test user
DATA_ITEST_USER = {
    'pkey': PKEY,
    'add': [
        ('textbox', 'uid', PKEY),
        ('textbox', 'givenname', 'itest-user-name'),
        ('textbox', 'sn', 'itest-user-surname'),
        ('password', 'userpassword', PASSWD_ITEST_USER),
        ('password', 'userpassword2', PASSWD_ITEST_USER),
    ]
}

# used for checking login form after click Cancel on 'reset' view
FILLED_LOGIN_FORM = {
    # structure of rows
    # label_name, label_text,
    # required, editable,
    # input_type, input_name,
    # input_text, placeholder
    'rows': [
        ('username', 'Username', True, True, 'text', 'username',
         PKEY, 'Username'),
        ('password', 'Password', True, True, 'password', 'password',
         PASSWD_ITEST_USER, 'Password or Password+One-Time-Password'),
    ],
    # structure of buttons
    # button_name, button_title
    'buttons': [
        ('cert_auth', 'Log in using personal certificate'),
        ('sync', 'Sync OTP Token'),
        ('login', 'Log in'),
    ],
    'required_msg': [
        ('Username: Required field',),
        ('Password: Required field',),
    ],
}

# used for checking 'reset_and_login' view
RESET_AND_LOGIN_FORM = {
    # structure of rows
    # label_name, label_text,
    # required, editable,
    # input_type, input_name,
    # input_text, placeholder
    'rows': [
        ('username_r', 'Username', False, False, None, 'username_r',
         PKEY, None),
        ('current_password', 'Current Password', False, True, 'password',
         'current_password', '', 'Current Password'),
        ('otp', 'OTP', False, True, 'password', 'otp', '',
         'One-Time-Password'),
        ('new_password', 'New Password', True, True, 'password',
         'new_password', '', 'New Password'),
        ('verify_password', 'Verify Password', True, True, 'password',
         'verify_password', '', 'New Password'),
    ],
    # structure of buttons
    # button_name, button_title
    'buttons': [
        ('cancel', 'Cancel'),
        ('reset_and_login', 'Reset Password and Log in'),
    ],
    'required_msg': [
        ('New Password: Required field',),
        ('Verify Password: Required field',),
    ],
}

# used for checking 'reset' view
RESET_PASSWORD_FORM = {
    # structure of rows
    # label_name, label_text,
    # required, editable,
    # input_type, input_name,
    # input_text, placeholder
    'rows': [
        ('username', 'Username', True, True, 'text', 'username', '',
         'Username'),
        ('current_password', 'Current Password', True, True, 'password',
         'current_password', '', 'Current Password'),
        ('otp', 'OTP', False, True, 'password', 'otp', '',
         'One-Time-Password'),
        ('new_password', 'New Password', True, True, 'password',
         'new_password', '', 'New Password'),
        ('verify_password', 'Verify Password', True, True, 'password',
         'verify_password', '', 'New Password'),
    ],
    # structure of buttons
    # button_name, button_title
    'buttons': [
        ('reset', 'Reset Password'),
    ],
    'required_msg': [
        ('Username: Required field',),
        ('Current Password: Required field',),
        ('New Password: Required field',),
        ('Verify Password: Required field',),
    ],
}


# used for checking empty 'login' view
EMPTY_LOGIN_FORM = {
    # structure of rows
    # label_name, label_text,
    # required, editable,
    # input_type, input_name,
    # input_text, placeholder
    'rows': [
        ('username', 'Username', False, True, 'text', 'username', '',
         'Username'),
        ('password', 'Password', False, True, 'password', 'password', '',
         'Password or Password+One-Time-Password'),
    ],
    # structure of buttons
    # button_name, button_title
    'buttons': [
        ('cert_auth', 'Log in using personal certificate'),
        ('sync', 'Sync OTP Token'),
        ('login', 'Log in'),
    ],
    'required_msg': [
        ('Authentication with Kerberos failed',),
    ],
}

# used for checking 'login' view
LOGIN_FORM = {
    # structure of rows
    # label_name, label_text,
    # required, editable,
    # input_type, input_name,
    # input_text, placeholder
    'rows': [
        ('username', 'Username', True, True, 'text', 'username', PKEY,
         'Username'),
        ('password', 'Password', True, True, 'password', 'password', '',
         'Password or Password+One-Time-Password'),
    ],
    # structure of buttons
    # button_name, button_title
    'buttons': [
        ('cert_auth', 'Log in using personal certificate'),
        ('sync', 'Sync OTP Token'),
        ('login', 'Log in'),
    ],
    'required_msg': [
        ('Password: Required field',),
    ],
}



# Authors:
#   Alexander Bokovoy <abokovoy@redhat.com>
#
# Copyright (C) 2011  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ipaclient.frontend import CommandOverride
from ipalib.plugable import Registry

import six

if six.PY3:
    unicode = str

register = Registry()


@register(override=True, no_fail=True)
class hbactest(CommandOverride):
    def output_for_cli(self, textui, output, *args, **options):
        """
        Command.output_for_cli() uses --all option to decide whether to print detailed output.
        We use --detail to allow that, thus we need to redefine output_for_cli().
        """
        # Note that we don't actually use --detail below to see if details need
        # to be printed as our execute() method will return None for corresponding
        # entries and None entries will be skipped.
        for o in self.output:
            if o == 'value':
                continue
            outp = self.output[o]
            if 'no_display' in outp.flags:
                continue
            result = output[o]
            if isinstance(result, (list, tuple)):
                textui.print_attribute(unicode(outp.doc), result, '%s: %s', 1, True)
            elif isinstance(result, unicode):
                if o == 'summary':
                    textui.print_summary(result)
                else:
                    textui.print_indented(result)

        # Propagate integer value for result. It will give proper command line result for scripts
        return int(not output['value'])
